from dataclasses import dataclass
from dataclasses import field as dataclass_field
from typing import Any, ClassVar, Dict, List, Optional, Tuple

from ..constants import (
    DESC_FIELD_NAME,
    DESC_FIELD_NAME_MD,
    FIELD_ID_STRING_SCHEMA,
    RELATION_STEPS_SCHEMA,
    SOURCE_ID_STRING_SCHEMA,
    VALUE_CAST_ENUM,
    schema_meta,
    schema_omit,
)
from ..doc_texts import SOURCE_FIELD_EXTRACT_DESC, SOURCE_FIELD_EXTRACT_MD
from .lookup_bind_relation import InlineRelationConfig


@dataclass(frozen=True)
class SourceFieldConfig:
    SCHEMA_NAME: ClassVar[str] = "source_field"
    """源字段配置对象在 `YAML` 中的节点名称."""

    SCHEMA_ALL_OF: ClassVar[List[Dict[str, Any]]] = [
        {"not": {"required": ["call_by"]}},
        {"not": {"required": ["field"]}},
    ]
    """用于约束:源字段配置中禁止声明 `call_by` 与历史 `field`."""

    field_id: str = dataclass_field(default="", metadata=schema_omit())
    """字段标识(内部字段;由外层映射键提供)."""

    source: str = dataclass_field(
        default="",
        metadata=schema_meta(
            schema={
                **SOURCE_ID_STRING_SCHEMA,
                "description": "字段来源的 source_id (例: source: orders)",
                "markdownDescription": (
                    "字段来源的 `source_id`.\n\n"
                    "- 在 `main_source.fields` / `sources.<id>.fields` 中可省略\n"
                    "- 若显式提供, 必须与容器 `source_id` 一致"
                ),
            }
        ),
    )
    """字段来源的 `source_id`(可选;在容器内通常可省略)."""

    extract: Optional[str] = dataclass_field(
        default=None,
        metadata=schema_meta(
            schema={
                "type": "string",
                "minLength": 1,
                "description": SOURCE_FIELD_EXTRACT_DESC,
                "markdownDescription": SOURCE_FIELD_EXTRACT_MD,
                "examples": [
                    "review_status",
                    "CustomerMark.clearn_reason_level",
                    "[1].clearn_reason_level",
                    '["a.b"].x',
                ],
            }
        ),
    )
    """可选:字段提取表达式(缺省时等于字段键/`field_id`)."""

    name: str = dataclass_field(default="", metadata=schema_meta(desc=DESC_FIELD_NAME, md=DESC_FIELD_NAME_MD))
    """字段展示名称(可选)."""

    relation: Optional[InlineRelationConfig] = dataclass_field(
        default=None,
        metadata=schema_meta(
            schema={
                "oneOf": [
                    {
                        **FIELD_ID_STRING_SCHEMA,
                        "description": "relation 引用(字符串 ref),示例: relation: orders_to_customers",
                        "markdownDescription": (
                            "relation 引用(字符串 ref).\n\n"
                            "- 形式: `relation: <relation_id>`\n"
                            "- 表示引用 `relations.<relation_id>`\n"
                            "- full validate 会校验 `relations.<relation_id>` 存在"
                        ),
                    },
                    {
                        "type": "object",
                        "required": ["steps"],
                        "properties": {"steps": RELATION_STEPS_SCHEMA},
                        "additionalProperties": False,
                        "description": "内联 steps 或 alias 引用(需先定义),示例: relation: {steps: [...]}",
                        "markdownDescription": (
                            "关系路径引用(内联 steps 或 YAML alias; alias 需先在 relations 定义).\n\n"
                            "- `relation: *orders_to_customers`\n"
                            "- `relation: {steps: [...]}`\n"
                            "- steps 必须从 `main_source` 开始, 以当前 `source` 结束\n"
                            "- 若 `source` 非 `main_source`, 且未显式 relation, 需确保 `relations` 中存在唯一链路"
                        ),
                    },
                ]
            },
            desc=(
                "关系路径(支持 string ref / steps 对象 / YAML alias; alias 需先定义),"
                "表示从 main_source 到当前字段 source 的等值关联链 (例: relation: orders_to_customers)"
            ),
        ),
    )
    """可选:关系路径引用(内联 `steps` 或 `YAML` 别名)."""

    value_cast: Optional[str] = dataclass_field(
        default=None,
        metadata=schema_meta(
            desc="字段值转换(仅源字段),用于写入上下文/输出前的类型调整",
            md=("字段值转换(仅源字段).\n\n- `auto`: 自动转换\n- `int`: 转为 int\n- `str`: 转为 str"),
            choices=VALUE_CAST_ENUM,
            examples=["auto"],
        ),
    )
    """可选:字段值类型转换策略(仅源字段)."""


@dataclass(frozen=True)
class DerivedFieldConfig:
    SCHEMA_NAME: ClassVar[str] = "derived_field"
    """派生字段配置对象在 `YAML` 中的节点名称."""

    field_id: str = dataclass_field(default="", metadata=schema_omit())
    """字段标识(内部字段;由外层映射键提供)."""

    name: str = dataclass_field(default="", metadata=schema_meta(desc=DESC_FIELD_NAME, md=DESC_FIELD_NAME_MD))
    """字段展示名称(可选)."""

    compute: Optional[str] = dataclass_field(
        default=None,
        metadata=schema_meta(
            desc="派生字段计算表达式(使用 field_id 作为变量名)",
            md=("派生字段计算表达式(与 `call_by` 互斥).\n\n- 使用 `field_id` 作为变量名\n- 必填, 不能为空"),
            examples=["revenue - cost"],
        ),
    )
    """派生字段计算表达式(与 `call_by` 互斥)."""

    call_by: Optional[str] = dataclass_field(
        default=None,
        metadata=schema_meta(
            desc="派生字段函数调用(函数引用 + 参数列表),与 compute 互斥",
            md=(
                "派生字段函数调用(与 `compute` 互斥).\n\n"
                "- 语法: `reference(args...)`\n"
                "- `reference` 形式与 `loader` 引用一致,支持:\n"
                "  - 绝对引用: `module.path.function` / `module.path:function` / `module.path:obj.method`\n"
                "  - 相对引用: 以 `.` / `..` 开头的 module path(相对 YAML 文件所在目录)\n"
                "- 相对引用会在运行期先归一化为绝对引用,并继续受 allowlist(allowed_modules/allowed_functions) 约束\n"
                "- 支持位置参数与 kwargs\n"
                "- Python 字面量: `1`/`1.5`/`'ok'`/`True`/`False`/`None`\n"
                "- 上下文引用: `$ctx` 或 `$ctx.<attr>`\n"
                "- 可用 ctx 属性: `row_id`/`batch_num`/`field_id`/`deps`/`values`"
            ),
            examples=[
                "myapp.enums:get_status_text(status)",
                "myapp.enums:get_status_text(status=status, ctx=$ctx)",
                ".helpers:to_text(status)",
            ],
        ),
    )
    """派生字段函数调用(与 `compute` 互斥)."""

    depends_on: Tuple[str, ...] = dataclass_field(default_factory=tuple, metadata=schema_omit())
    """依赖字段标识列表(内部字段;解析后填充)."""



# region imports

import threading
from typing import Any, Callable, Dict, Hashable, List, Optional, TypeVar, Union

from ..events.catalog import (
    EVENT_BATCH_END,
    EVENT_BATCH_START,
    EVENT_COLUMN_WRITE,
    EVENT_DIAGNOSTIC_WARNING,
    EVENT_ERROR,
    EVENT_FIELD_COMPUTE,
    EVENT_FIELD_SLIM,
    EVENT_LOADER_CALL,
    EVENT_LOADER_RETRY,
    EVENT_LOADER_SLIM,
    EVENT_PIPELINE_END,
    EVENT_PIPELINE_START,
    EVENT_RELATION_LOOKUP,
    EVENT_ROW_RELEASE,
    EVENT_ROW_WRITE,
    EVENT_STAGE_SPAN,
)
from ..events.event import Event
from ..events.events import (
    BatchEndEvent,
    BatchStartEvent,
    ColumnWriteEvent,
    DiagnosticWarningEvent,
    ErrorEvent,
    FieldComputeEvent,
    FieldSlimEvent,
    LoaderCallEvent,
    LoaderRetryEvent,
    LoaderSlimEvent,
    PipelineEndEvent,
    PipelineStartEvent,
    RelationLookupEvent,
    RowReleaseEvent,
    RowWriteEvent,
    StageSpanEvent,
)
from ..hooks.base import HookManager, IExecutionHook
from ..typedefs import RelationLookupResult
from .components import split_components
from .manager import ObserverManager
from .observer import Observer

# endregion

_PayloadT = TypeVar("_PayloadT")


class InstrumentationHub:
    """统一的观测枢纽: 负责钩子与观测器.

    执行热路径应尽量只调用一次该枢纽,并使用 `wants(event_type)` 来门控昂贵的载荷构造.
    """

    hook_manager: HookManager
    observer_manager: ObserverManager
    _lock: "threading.RLock"
    _diagnostic_warning_emitted: bool

    def __init__(
        self,
        hook_manager: Optional[HookManager] = None,
        observer_manager: Optional[ObserverManager] = None,
    ) -> None:
        self.hook_manager = hook_manager or HookManager()
        self.observer_manager = observer_manager or ObserverManager()
        self._lock = threading.RLock()
        self._diagnostic_warning_emitted = False

    def __getstate__(self) -> Dict[str, Any]:
        state = dict(self.__dict__)
        state.pop("_lock", None)
        return state

    def __setstate__(self, state: Dict[str, Any]) -> None:
        self.__dict__.update(state)
        self._lock = threading.RLock()
        if not hasattr(self, "_diagnostic_warning_emitted"):
            self._diagnostic_warning_emitted = False

    def register(self, subscriber: Union[Observer, IExecutionHook]) -> None:
        observers, hooks = split_components([subscriber])
        if observers:
            self.observer_manager.register(observers[0])
            return
        self.hook_manager.register(hooks[0])

    def unregister(self, subscriber: Union[Observer, IExecutionHook]) -> bool:
        observers, hooks = split_components([subscriber])
        if observers:
            return self.observer_manager.unregister(observers[0])
        return self.hook_manager.unregister(hooks[0])

    def clear(self) -> None:
        self.hook_manager.clear()
        self.observer_manager.clear()

    def wants(self, event_type: str) -> bool:
        # `wants` 为 O(1) 检查;用于热路径在无人订阅时避免构造载荷/关键字参数.
        return self.hook_manager.wants(event_type) or self.observer_manager.wants(event_type)

    def _emit_assume_wanted(
        self,
        event_type: str,
        payload: Any,
        meta: Optional[Dict[str, Any]] = None,
    ) -> Optional[Event]:
        self.hook_manager.emit_typed(event_type, payload)

        if not (self.observer_manager.wants(event_type) or self.hook_manager.wants_on_event(event_type)):
            return None

        event = self.observer_manager.emit_event(event_type, payload, meta=meta)
        if self.observer_manager.mode != "capture":
            self.hook_manager.emit_on_event(event)
        return event

    def emit_lazy(
        self,
        event_type: str,
        payload_factory: Callable[[], _PayloadT],
        meta: Optional[Dict[str, Any]] = None,
    ) -> Optional[Event]:
        if not self.wants(event_type):
            return None
        payload = payload_factory()
        return self._emit_assume_wanted(event_type, payload, meta=meta)

    def emit(
        self,
        event_type: str,
        payload: Any,
        meta: Optional[Dict[str, Any]] = None,
    ) -> Optional[Event]:
        """向类型化钩子、观测器以及订阅了 `hook.on_event(Event)` 的监听方发送事件.

        当事件会被发送到观测器/`on_event` 路径时,返回构建好的 `Event` 包装;否则返回 `None`.
        """
        if not self.wants(event_type):
            return None
        return self._emit_assume_wanted(event_type, payload, meta=meta)

    def emit_recorded_event(self, event: Event) -> None:
        """回放此前记录的事件(例如来自并行捕获模式)."""
        self.observer_manager.emit(event)
        if self.observer_manager.mode != "capture":
            self.hook_manager.emit_on_event(event)

    # ---- 类型化辅助方法(执行热路径优先使用) ----

    def emit_pipeline_start(self, targets: List[str], batch_size: Optional[int]) -> None:
        if not self.wants(EVENT_PIPELINE_START):
            return
        _ = self._emit_assume_wanted(EVENT_PIPELINE_START, PipelineStartEvent(targets, batch_size))

    def emit_pipeline_end(self, total_batches: int, total_duration: float) -> None:
        if not self.wants(EVENT_PIPELINE_END):
            return
        _ = self._emit_assume_wanted(EVENT_PIPELINE_END, PipelineEndEvent(total_batches, total_duration))

    def emit_batch_start(self, batch_num: int, row_ids: List[Any]) -> None:
        if not self.wants(EVENT_BATCH_START):
            return
        _ = self._emit_assume_wanted(EVENT_BATCH_START, BatchStartEvent(batch_num, row_ids))

    def emit_batch_end(self, batch_num: int, duration: float) -> None:
        if not self.wants(EVENT_BATCH_END):
            return
        _ = self._emit_assume_wanted(EVENT_BATCH_END, BatchEndEvent(batch_num, duration))

    def emit_loader_call(
        self,
        loader_name: str,
        params: Dict[str, Any],
        result: Any,
        duration: float,
        *,
        batch_num: Optional[int] = None,
        cache_status: Optional[str] = None,
        cache_scope: Optional[str] = None,
        lookup_key_count: Optional[int] = None,
        field_keys: Optional[List[str]] = None,
        meta: Optional[Dict[str, Any]] = None,
    ) -> None:
        # 类型化钩子的 `loader_result_policy` 可能与观测器/事件路径不同.
        if self.hook_manager.wants_typed(EVENT_LOADER_CALL):
            self.hook_manager.trigger_loader_call(
                loader_name=loader_name,
                params=params,
                result=result,
                duration=duration,
                batch_num=batch_num,
                cache_status=cache_status,
                cache_scope=cache_scope,
                lookup_key_count=lookup_key_count,
                field_keys=field_keys,
            )

        if not (self.observer_manager.wants(EVENT_LOADER_CALL) or self.hook_manager.wants_on_event(EVENT_LOADER_CALL)):
            return

        payload = result
        if self.observer_manager.loader_result_policy != "full":
            if self.observer_manager.loader_result_policy == "none":
                payload = None
            elif self.observer_manager.loader_result_policy == "summary":
                payload = self.observer_manager.summarize_result(result)
            elif self.observer_manager.loader_result_policy == "sample":
                payload = self.observer_manager.sample_result(result)

        event_payload = LoaderCallEvent(
            loader_name=loader_name,
            params=params,
            result=payload,
            duration=duration,
            batch_num=batch_num,
            cache_status=cache_status,
            cache_scope=cache_scope,
            lookup_key_count=lookup_key_count,
            field_keys=field_keys,
        )
        event = self.observer_manager.emit_event(EVENT_LOADER_CALL, event_payload, meta=meta)
        if self.observer_manager.mode != "capture":
            self.hook_manager.emit_on_event(event)

    def emit_loader_retry(
        self,
        *,
        loader_name: str,
        callsite: str,
        attempt_num: int,
        max_attempts: int,
        elapsed_seconds: float,
        sleep_seconds: float,
        error_type: str,
        error_message: Optional[str],
        batch_num: Optional[int] = None,
        meta: Optional[Dict[str, Any]] = None,
    ) -> None:
        if not self.wants(EVENT_LOADER_RETRY):
            return
        _ = self._emit_assume_wanted(
            EVENT_LOADER_RETRY,
            LoaderRetryEvent(
                loader_name=loader_name,
                callsite=callsite,
                attempt_num=int(attempt_num),
                max_attempts=int(max_attempts),
                elapsed_seconds=float(elapsed_seconds),
                sleep_seconds=float(sleep_seconds),
                error_type=str(error_type),
                error_message=error_message,
                batch_num=batch_num,
            ),
            meta=meta,
        )

    def emit_field_compute(
        self,
        field_key: str,
        row_id: Hashable,
        dependencies: Dict[str, Any],
        result: Any,
        meta: Optional[Dict[str, Any]] = None,
    ) -> None:
        if not self.wants(EVENT_FIELD_COMPUTE):
            return
        _ = self._emit_assume_wanted(EVENT_FIELD_COMPUTE, FieldComputeEvent(field_key, row_id, dependencies, result), meta=meta)

    def emit_error(self, error: Exception, context: Dict[str, Any], meta: Optional[Dict[str, Any]] = None) -> None:
        if not self.wants(EVENT_ERROR):
            return
        _ = self._emit_assume_wanted(EVENT_ERROR, ErrorEvent(error, context), meta=meta)

    def emit_diagnostic_warning(
        self,
        message: str,
        source_id: str,
        field_id: str,
        lookup_key: Any,
        row_id: Hashable,
        *,
        sample_once: bool = False,
        meta: Optional[Dict[str, Any]] = None,
    ) -> None:
        if sample_once:
            with self._lock:
                if self._diagnostic_warning_emitted:
                    return
                self._diagnostic_warning_emitted = True

        hooks_want = self.hook_manager.wants_typed(EVENT_DIAGNOSTIC_WARNING)
        event_want = self.observer_manager.wants(EVENT_DIAGNOSTIC_WARNING) or self.hook_manager.wants_on_event(EVENT_DIAGNOSTIC_WARNING)

        if not hooks_want and not event_want:
            if self.hook_manager.fallback_logger_enabled or self.observer_manager.fallback_logger_enabled:
                # 为保持一致性,这里使用 `ObserverManager` 的日志记录器.
                self.observer_manager.emit_diagnostic_warning(
                    message=message,
                    source_id=source_id,
                    field_id=field_id,
                    lookup_key=lookup_key,
                    row_id=row_id,
                    sample_once=False,
                )
            return

        payload = DiagnosticWarningEvent(
            message=message,
            source_id=source_id,
            field_id=field_id,
            lookup_key=lookup_key,
            row_id=row_id,
        )

        if hooks_want:
            self.hook_manager.emit_typed(EVENT_DIAGNOSTIC_WARNING, payload)

        if not event_want:
            return

        event = self.observer_manager.emit_event(EVENT_DIAGNOSTIC_WARNING, payload, meta=meta)
        if self.observer_manager.mode != "capture":
            self.hook_manager.emit_on_event(event)

    def emit_field_slim(
        self,
        field_key: str,
        reason: str,
        batch_num: int,
        remaining_fields: int,
        meta: Optional[Dict[str, Any]] = None,
    ) -> None:
        if not self.wants(EVENT_FIELD_SLIM):
            return
        _ = self._emit_assume_wanted(
            EVENT_FIELD_SLIM,
            FieldSlimEvent(field_key, reason, batch_num, remaining_fields),
            meta=meta,
        )

    def emit_row_write(
        self,
        row_id: Hashable,
        field_count: int,
        batch_num: int,
        row_index: int,
        meta: Optional[Dict[str, Any]] = None,
    ) -> None:
        if not self.wants(EVENT_ROW_WRITE):
            return
        _ = self._emit_assume_wanted(EVENT_ROW_WRITE, RowWriteEvent(row_id, field_count, batch_num, row_index), meta=meta)

    def emit_row_release(
        self,
        row_id: Hashable,
        released_fields: List[str],
        retained_fields: List[str],
        batch_num: int,
        meta: Optional[Dict[str, Any]] = None,
    ) -> None:
        if not self.wants(EVENT_ROW_RELEASE):
            return
        _ = self._emit_assume_wanted(
            EVENT_ROW_RELEASE,
            RowReleaseEvent(row_id, released_fields, retained_fields, batch_num),
            meta=meta,
        )

    def emit_loader_slim(
        self,
        loader_name: str,
        original_keys: int,
        extracted_fields: List[str],
        batch_num: int,
        meta: Optional[Dict[str, Any]] = None,
    ) -> None:
        if not self.wants(EVENT_LOADER_SLIM):
            return
        _ = self._emit_assume_wanted(
            EVENT_LOADER_SLIM,
            LoaderSlimEvent(loader_name, original_keys, extracted_fields, batch_num),
            meta=meta,
        )

    def emit_column_write(
        self,
        field_key: str,
        row_count: int,
        batch_num: int,
        meta: Optional[Dict[str, Any]] = None,
    ) -> None:
        if not self.wants(EVENT_COLUMN_WRITE):
            return
        _ = self._emit_assume_wanted(EVENT_COLUMN_WRITE, ColumnWriteEvent(field_key, row_count, batch_num), meta=meta)

    def emit_relation_lookup(
        self,
        field_key: str,
        row_id: Hashable,
        fk_raw: Any,
        fk_normalized: Any,
        target_source: str,
        result: RelationLookupResult,
        fk_type: Optional[str] = None,
        expected_type: Optional[str] = None,
        error_message: Optional[str] = None,
        meta: Optional[Dict[str, Any]] = None,
    ) -> None:
        if not self.wants(EVENT_RELATION_LOOKUP):
            return
        payload = RelationLookupEvent(
            field_key=field_key,
            row_id=row_id,
            fk_raw=fk_raw,
            fk_normalized=fk_normalized,
            target_source=target_source,
            result=result,
            fk_type=fk_type,
            expected_type=expected_type,
            error_message=error_message,
        )
        _ = self._emit_assume_wanted(EVENT_RELATION_LOOKUP, payload, meta=meta)

    def emit_stage_span(self, stage: str, batch_num: int, duration: float, meta: Optional[Dict[str, Any]] = None) -> None:
        if not self.wants(EVENT_STAGE_SPAN):
            return
        _ = self._emit_assume_wanted(EVENT_STAGE_SPAN, StageSpanEvent(stage=stage, batch_num=batch_num, duration=duration), meta=meta)


__all__ = [
    "InstrumentationHub",
]



"""`YAML` `DSL` 官方入口.

普通用户建议仅从此处导入最常用的运行入口与运行期契约,避免误用内部实现路径.

更底层的能力(解析/校验/编译/可视化/自省等)请从子模块显式导入,例如:
- `scalim.dsl.by_yaml.config_parsing.*`
- `scalim.dsl.by_yaml.runtime.*`
"""

from typing import TYPE_CHECKING, Any

from ...vendor.compact.importlibx import import_module
from .runtime.contracts import (
    UNSET,
    Compilation,
    OutputOverrides,
    RunOptions,
    RunOverrides,
    RunResult,
)

if TYPE_CHECKING:
    from .runtime.entrypoints import compile as compile  # noqa: A004
    from .runtime.entrypoints import run as run
    from .runtime.workflow_entrypoints import run_workflow as run_workflow
else:

    def compile(*args: Any, **kwargs: Any) -> Compilation:  # noqa: A001
        entrypoints = import_module("scalim.dsl.by_yaml.runtime.entrypoints")
        return entrypoints.compile(*args, **kwargs)

    def run(*args: Any, **kwargs: Any) -> RunResult:
        entrypoints = import_module("scalim.dsl.by_yaml.runtime.entrypoints")
        return entrypoints.run(*args, **kwargs)

    def run_workflow(*args: Any, **kwargs: Any) -> Any:
        entrypoints = import_module("scalim.dsl.by_yaml.runtime.workflow_entrypoints")
        return entrypoints.run_workflow(*args, **kwargs)


__all__ = (
    "UNSET",
    "Compilation",
    "OutputOverrides",
    "RunOptions",
    "RunOverrides",
    "RunResult",
    "compile",
    "run",
    "run_workflow",
)



# region imports

import threading
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional, Set, Tuple

from ..events.event import Event
from ..events.events import (
    BatchEndEvent,
    BatchStartEvent,
    ColumnWriteEvent,
    DiagnosticWarningEvent,
    ErrorEvent,
    FieldComputeEvent,
    FieldSlimEvent,
    LoaderCallEvent,
    LoaderSlimEvent,
    PipelineEndEvent,
    PipelineStartEvent,
    RowReleaseEvent,
    RowWriteEvent,
)
from ..vendor.compact.typing_extensionsx import override
from ._internal.common import HOOK_RAISED_EXCEPTION_WARNING, HOOK_TYPED_DISPATCH_MAP
from ._internal.manager_events import HookManagerEventMixin
from ._internal.manager_registry import HookManagerRegistryMixin
from ._internal.manager_state import HookManagerStateMixin
from ._internal.manager_subscriptions import HookManagerSubscriptionMixin
from .dispatch import HookDispatchStrategy

# endregion

_TypedHandlerPair = Tuple["IExecutionHook", Callable[[Any], Any]]
_OnEventHandlerPair = Tuple["IExecutionHook", Callable[[Event], Any]]


class IExecutionHook(ABC):
    """执行钩子接口."""

    @abstractmethod
    def on_pipeline_start(self, event: PipelineStartEvent) -> None:
        """当管线开始时调用."""

    @abstractmethod
    def on_pipeline_end(self, event: PipelineEndEvent) -> None:
        """当管线结束时调用."""

    @abstractmethod
    def on_batch_start(self, event: BatchStartEvent) -> None:
        """当批次开始时调用."""

    @abstractmethod
    def on_batch_end(self, event: BatchEndEvent) -> None:
        """当批次结束时调用."""

    @abstractmethod
    def on_loader_call(self, event: LoaderCallEvent) -> None:
        """当加载函数被调用时调用."""

    @abstractmethod
    def on_field_compute(self, event: FieldComputeEvent) -> None:
        """当字段被计算时调用"""

    @abstractmethod
    def on_error(self, event: ErrorEvent) -> None:
        """当发生错误时调用"""

    @abstractmethod
    def on_diagnostic_warning(self, event: DiagnosticWarningEvent) -> None:
        """当发生诊断告警时调用"""

    @abstractmethod
    def on_field_slim(self, event: FieldSlimEvent) -> None:
        """当字段从上下文中删除时调用 (FR022)"""

    @abstractmethod
    def on_row_write(self, event: RowWriteEvent) -> None:
        """当行被写入行式输出端时调用 (FR023)."""

    @abstractmethod
    def on_row_release(self, event: RowReleaseEvent) -> None:
        """当行的内存被释放时调用 (FR023)"""

    @abstractmethod
    def on_loader_slim(self, event: LoaderSlimEvent) -> None:
        """当加载结果被压缩时调用 (FR022)."""

    @abstractmethod
    def on_column_write(self, event: ColumnWriteEvent) -> None:
        """当列被写入列式输出端时调用 (FR023)."""


Hook = IExecutionHook


class BaseHook(IExecutionHook):
    """带有空操作方法的基础钩子实现."""

    event_types: Optional[Set[str]] = None

    def on_event(self, event: Event) -> None:
        """统一事件回调(用于订阅事件目录中的全部事件)."""
        _ = event

    @override
    def on_pipeline_start(self, event: PipelineStartEvent) -> None:
        """空操作实现"""

    @override
    def on_pipeline_end(self, event: PipelineEndEvent) -> None:
        """空操作实现"""

    @override
    def on_batch_start(self, event: BatchStartEvent) -> None:
        """空操作实现"""

    @override
    def on_batch_end(self, event: BatchEndEvent) -> None:
        """空操作实现"""

    @override
    def on_loader_call(self, event: LoaderCallEvent) -> None:
        """空操作实现"""

    @override
    def on_field_compute(self, event: FieldComputeEvent) -> None:
        """空操作实现"""

    @override
    def on_error(self, event: ErrorEvent) -> None:
        """空操作实现"""

    @override
    def on_diagnostic_warning(self, event: DiagnosticWarningEvent) -> None:
        """空操作实现"""

    @override
    def on_field_slim(self, event: FieldSlimEvent) -> None:
        """空操作实现"""

    @override
    def on_row_write(self, event: RowWriteEvent) -> None:
        """空操作实现"""

    @override
    def on_row_release(self, event: RowReleaseEvent) -> None:
        """空操作实现"""

    @override
    def on_loader_slim(self, event: LoaderSlimEvent) -> None:
        """空操作实现"""

    @override
    def on_column_write(self, event: ColumnWriteEvent) -> None:
        """空操作实现"""


class HookManager(HookManagerStateMixin, HookManagerSubscriptionMixin, HookManagerRegistryMixin, HookManagerEventMixin):
    """钩子管理器 - 管理并触发所有钩子.

    注意: 请使用 `register`/`unregister`/`clear` 管理钩子,不要直接修改 `hooks` 列表,否则 `fastpath` 缓存可能失效.
    """

    hooks: List[IExecutionHook]
    _has_hooks: bool
    _typed_handlers_by_event_type: Dict[str, Tuple[_TypedHandlerPair, ...]]
    _on_event_handlers_by_event_type: Dict[str, Tuple[_OnEventHandlerPair, ...]]
    debug_mode: bool
    fallback_logger_enabled: bool
    loader_result_policy: str
    loader_result_sample_size: int
    _lock: "threading.RLock"
    _diagnostic_warning_emitted: bool
    _dispatch_strategy: HookDispatchStrategy
    _base_hook_on_event: Optional[Callable[[BaseHook, Event], None]]
    _base_hook_typed_handlers: Dict[str, Optional[Callable[..., None]]]

    def __init__(
        self,
        enable_debugging: bool = False,  # noqa: FBT001, FBT002
        fallback_logger_enabled: bool = False,  # noqa: FBT001, FBT002
        loader_result_policy: str = "full",
        loader_result_sample_size: int = 5,
        dispatch_strategy: Optional[HookDispatchStrategy] = None,
    ) -> None:
        self.hooks = []
        self._has_hooks = False
        self._typed_handlers_by_event_type = {}
        self._on_event_handlers_by_event_type = {}
        self._base_hook_on_event = BaseHook.__dict__.get("on_event")
        self._base_hook_typed_handlers = {name: BaseHook.__dict__.get(name) for name in HOOK_TYPED_DISPATCH_MAP.values()}
        self.debug_mode = enable_debugging
        self.fallback_logger_enabled = fallback_logger_enabled
        self.loader_result_policy = self._normalize_loader_result_policy(loader_result_policy)
        self.loader_result_sample_size = max(1, loader_result_sample_size)
        self._lock = threading.RLock()
        self._diagnostic_warning_emitted = False
        self._dispatch_strategy = dispatch_strategy or HookDispatchStrategy()

    @property
    def has_hooks(self) -> bool:
        return self._has_hooks

    @has_hooks.setter
    def has_hooks(self, value: bool) -> None:
        self._has_hooks = value

    @property
    def typed_handlers_by_event_type(self) -> Dict[str, Tuple[_TypedHandlerPair, ...]]:
        return self._typed_handlers_by_event_type

    @typed_handlers_by_event_type.setter
    def typed_handlers_by_event_type(self, value: Dict[str, Tuple[_TypedHandlerPair, ...]]) -> None:
        self._typed_handlers_by_event_type = value

    @property
    def on_event_handlers_by_event_type(self) -> Dict[str, Tuple[_OnEventHandlerPair, ...]]:
        return self._on_event_handlers_by_event_type

    @on_event_handlers_by_event_type.setter
    def on_event_handlers_by_event_type(self, value: Dict[str, Tuple[_OnEventHandlerPair, ...]]) -> None:
        self._on_event_handlers_by_event_type = value

    @property
    def lock(self) -> "threading.RLock":
        return self._lock

    @lock.setter
    def lock(self, value: "threading.RLock") -> None:
        self._lock = value

    @property
    def diagnostic_warning_emitted(self) -> bool:
        return self._diagnostic_warning_emitted

    @diagnostic_warning_emitted.setter
    def diagnostic_warning_emitted(self, value: bool) -> None:
        self._diagnostic_warning_emitted = value

    @property
    def dispatch_strategy(self) -> HookDispatchStrategy:
        return self._dispatch_strategy

    @dispatch_strategy.setter
    def dispatch_strategy(self, value: HookDispatchStrategy) -> None:
        self._dispatch_strategy = value

    @property
    def base_hook_on_event(self) -> Optional[Callable[[BaseHook, Event], None]]:
        return self._base_hook_on_event

    @base_hook_on_event.setter
    def base_hook_on_event(self, value: Optional[Callable[[BaseHook, Event], None]]) -> None:
        self._base_hook_on_event = value

    @property
    def base_hook_typed_handlers(self) -> Dict[str, Optional[Callable[..., None]]]:
        return self._base_hook_typed_handlers

    @base_hook_typed_handlers.setter
    def base_hook_typed_handlers(self, value: Dict[str, Optional[Callable[..., None]]]) -> None:
        self._base_hook_typed_handlers = value


__all__ = [
    "HOOK_RAISED_EXCEPTION_WARNING",
    "BaseHook",
    "Hook",
    "HookManager",
    "IExecutionHook",
]



import re
import sys
from dataclasses import dataclass
from http.server import BaseHTTPRequestHandler, HTTPServer
from pathlib import Path
from typing import List, Optional, Sequence, Set, Tuple, cast
from urllib.parse import unquote, urlparse

from ..vendor.compact.typing_extensionsx import override

DEFAULT_SCHEMA_SERVE_HOST = "0.0.0.0"  # noqa: S104
DEFAULT_SCHEMA_SERVE_PORT = 62831

DEFAULT_SCHEMA_TYPE = "demand"
DEFAULT_SCHEMA_PATH = "http://localhost:62831"

DEFAULT_MAX_SCAN_LINES = 10

_SCHEMA_TYPE_PATTERN = re.compile(r"^[a-z][a-z0-9_-]*$")

_INTELLIJ_SCHEMA_PATTERN = re.compile(r"^\s*#\s*\$schema\s*:\s*(?P<ref>.*)\s*$")
_YAML_LANGUAGE_SERVER_SCHEMA_PATTERN = re.compile(r"^\s*#\s*yaml-language-server\s*:\s*\$schema\s*=\s*(?P<ref>.*)\s*$")


def schema_dir() -> Path:
    return Path(__file__).resolve().parents[1] / "dsl" / "by_yaml" / "schema"


def list_schema_filenames(schema_dir_path: Path) -> List[str]:
    return sorted(p.name for p in schema_dir_path.glob("*.gen.json") if p.is_file())


def resolve_schema_ref(schema_type: str, schema_path: str) -> str:
    schema_type = (schema_type or "").strip() or DEFAULT_SCHEMA_TYPE
    if not _SCHEMA_TYPE_PATTERN.match(schema_type):
        msg = "Invalid schema type: {}".format(schema_type)
        raise ValueError(msg)

    schema_path = (schema_path or "").strip() or DEFAULT_SCHEMA_PATH
    schema_filename = "{}.gen.json".format(schema_type)

    if schema_path.endswith(".json"):
        return schema_path

    if schema_path.startswith(("http://", "https://")):
        base_url = schema_path.rstrip("/")
        return "{}/{}".format(base_url, schema_filename)

    base_dir = Path(schema_path)
    return str(base_dir / schema_filename)


def make_schema_modeline(schema_ref: str) -> str:
    return "# $schema: {}".format(schema_ref)


def _is_schema_modeline(line: str) -> bool:
    return bool(_INTELLIJ_SCHEMA_PATTERN.match(line) or _YAML_LANGUAGE_SERVER_SCHEMA_PATTERN.match(line))


def upsert_schema_modeline_text(
    text: str,
    *,
    schema_modeline: str,
    max_scan_lines: int = DEFAULT_MAX_SCAN_LINES,
) -> Tuple[str, bool]:
    newline = "\r\n" if "\r\n" in text else "\n"
    ends_with_newline = text.endswith("\n")

    lines = text.splitlines()
    scan_limit = min(len(lines), max_scan_lines)

    modeline_indices: List[int] = []
    for idx in range(scan_limit):
        line = lines[idx]
        stripped = line.strip()

        if _is_schema_modeline(line):
            modeline_indices.append(idx)
            continue

        if not stripped:
            continue

        if stripped.startswith("#"):
            continue

        break

    changed = False
    if modeline_indices:
        first_idx = modeline_indices[0]
        if lines[first_idx] != schema_modeline:
            lines[first_idx] = schema_modeline
            changed = True
        for idx in reversed(modeline_indices[1:]):
            _ = lines.pop(idx)
            changed = True
    else:
        lines = [schema_modeline, "", *lines]
        changed = True

    new_text = newline.join(lines)
    if ends_with_newline:
        new_text += newline

    return new_text, changed


@dataclass
class UpsertResult:
    path: Path
    changed: bool
    error: Optional[str] = None


def upsert_schema_modeline_file(
    path: Path,
    *,
    schema_modeline: str,
    max_scan_lines: int = DEFAULT_MAX_SCAN_LINES,
) -> UpsertResult:
    try:
        text = path.read_text(encoding="utf-8")
    except Exception as exc:  # noqa: BLE001
        return UpsertResult(path=path, changed=False, error="Failed to read: {}".format(exc))

    new_text, changed = upsert_schema_modeline_text(text, schema_modeline=schema_modeline, max_scan_lines=max_scan_lines)
    if not changed:
        return UpsertResult(path=path, changed=False)

    try:
        _ = path.write_text(new_text, encoding="utf-8")
    except Exception as exc:  # noqa: BLE001
        return UpsertResult(path=path, changed=False, error="Failed to write: {}".format(exc))

    return UpsertResult(path=path, changed=True)


class _ReusableHTTPServer(HTTPServer):
    allow_reuse_address: bool = True


def _extract_single_filename(raw_path: str) -> Optional[str]:
    parsed = urlparse(raw_path)
    path = unquote(parsed.path or "")
    if path.startswith("/"):
        path = path[1:]
    if not path:
        return None
    if "/" in path or "\\" in path:
        return None
    return path


class SchemaHTTPServer(_ReusableHTTPServer):
    schema_dir_path: Path
    allowed_names: Set[str]

    def __init__(
        self,
        server_address: Tuple[str, int],
        handler_cls: type,
        *,
        schema_dir_path: Path,
        allowed_names: Set[str],
    ) -> None:
        self.schema_dir_path = schema_dir_path
        self.allowed_names = allowed_names
        super().__init__(server_address, handler_cls)


class SchemaHandler(BaseHTTPRequestHandler):
    server_version: str = "scalim-cli-schema-serve"

    def do_GET(self) -> None:
        filename = self._resolve_schema_filename()
        if filename is None:
            self._send_not_found()
            return

        schema_path = self._schema_dir_path() / filename
        if not schema_path.exists() or not schema_path.is_file():
            self._send_not_found()
            return

        try:
            payload = schema_path.read_bytes()
        except Exception:  # noqa: BLE001
            self._send_not_found()
            return

        self.send_response(200)
        self.send_header("Content-Type", "application/json; charset=utf-8")
        self.send_header("Content-Length", str(len(payload)))
        self.end_headers()
        _ = self.wfile.write(payload)

    def do_HEAD(self) -> None:
        filename = self._resolve_schema_filename()
        if filename is None:
            self._send_not_found()
            return

        schema_path = self._schema_dir_path() / filename
        if not schema_path.exists() or not schema_path.is_file():
            self._send_not_found()
            return

        self.send_response(200)
        self.send_header("Content-Type", "application/json; charset=utf-8")
        self.end_headers()

    def do_POST(self) -> None:
        self._send_method_not_allowed()

    def do_PUT(self) -> None:
        self._send_method_not_allowed()

    def do_DELETE(self) -> None:
        self._send_method_not_allowed()

    @override
    def log_message(self, format: str, *args: object) -> None:  # noqa: A002
        message = format % args
        _ = sys.stderr.write("%s - - [%s] %s\n" % (self.address_string(), self.log_date_time_string(), message))

    def _schema_dir_path(self) -> Path:
        server = cast("SchemaHTTPServer", self.server)
        return server.schema_dir_path

    def _resolve_schema_filename(self) -> Optional[str]:
        server = cast("SchemaHTTPServer", self.server)
        filename = _extract_single_filename(self.path)
        if filename is None:
            return None
        if filename not in server.allowed_names:
            return None
        if not filename.endswith(".gen.json"):
            return None
        return filename

    def _send_not_found(self) -> None:
        self.send_response(404)
        self.end_headers()

    def _send_method_not_allowed(self) -> None:
        self.send_response(405)
        self.end_headers()


def create_schema_http_server(
    *,
    host: str,
    port: int,
    schema_dir_path: Optional[Path] = None,
) -> Tuple[SchemaHTTPServer, int, List[str]]:
    schema_dir_path = schema_dir_path or schema_dir()
    allowed_names = list_schema_filenames(schema_dir_path)
    server = SchemaHTTPServer(
        (host, port),
        SchemaHandler,
        schema_dir_path=schema_dir_path,
        allowed_names=set(allowed_names),
    )
    actual_port = int(server.server_address[1])
    return server, actual_port, allowed_names


def print_schema_serve_banner(*, host: str, port: int, schema_filenames: Sequence[str]) -> None:
    base_url = "http://localhost:{}".format(port)
    _ = sys.stdout.write("正在提供 `YAML DSL` 的 `JSON Schema` 文件: {}\n".format(schema_dir()))
    _ = sys.stdout.write("监听地址: http://{}:{}\n".format(host, port))
    _ = sys.stdout.write("推荐 `$schema` 基址: {}\n".format(base_url))
    _ = sys.stdout.write("可用的 `schema` 文件:\n")
    for name in schema_filenames:
        _ = sys.stdout.write("- {}/{}\n".format(base_url, name))
    if schema_filenames:
        _ = sys.stdout.write("头部示例:\n")
        _ = sys.stdout.write("{}\n".format(make_schema_modeline("{}/{}".format(base_url, schema_filenames[0]))))



import re
from dataclasses import dataclass
from typing import Tuple

_MODULE_PATH_RE = re.compile(r"^[.]*[A-Za-z_][A-Za-z0-9_]*(?:[.][A-Za-z_][A-Za-z0-9_]*)*$")
_REFERENCE_PARTS_COUNT = 2

REFERENCE_FORMAT_EXAMPLES = "`module.path:function` / `module.path:obj.method` / `module.path.function`"


class ReferenceSyntaxError(ValueError):
    pass


@dataclass(frozen=True)
class ParsedReference:
    reference: str
    module_path: str
    attr_path: Tuple[str, ...]
    style: str

    @property
    def entry_attr(self) -> str:
        return self.attr_path[0]


def parse_python_reference(reference: str) -> ParsedReference:
    raw = str(reference or "").strip()
    if not raw:
        msg = "引用不能为空"
        raise ReferenceSyntaxError(msg)
    if ":" in raw:
        return _parse_class_style(raw)
    return _parse_dotted_style(raw)


def is_valid_python_reference(reference: str) -> bool:
    try:
        _ = parse_python_reference(reference)
    except ReferenceSyntaxError:
        return False
    return True


def _parse_class_style(reference: str) -> ParsedReference:
    parts = reference.split(":")
    if len(parts) != _REFERENCE_PARTS_COUNT:
        msg = "类式引用 '{}' 非法;期望格式: `module.path:attr` 或 `module.path:obj.method`".format(reference)
        raise ReferenceSyntaxError(msg)

    module_path, attr_path = parts
    if module_path and not module_path.strip("."):
        msg = "引用 '{}' 中的相对模块路径 '{}' 非法: 前导点后缺少模块路径".format(reference, module_path)
        raise ReferenceSyntaxError(msg)
    if not module_path or _MODULE_PATH_RE.fullmatch(module_path) is None:
        msg = "引用 '{}' 的模块路径 '{}' 非法".format(reference, module_path)
        raise ReferenceSyntaxError(msg)

    attr_parts = tuple(attr_path.split("."))
    invalid = [part for part in attr_parts if not part or not part.isidentifier()]
    if invalid:
        msg = "引用 '{}' 的属性路径 '{}' 非法".format(reference, attr_path)
        raise ReferenceSyntaxError(msg)

    return ParsedReference(reference=reference, module_path=module_path, attr_path=attr_parts, style="class")


def _parse_dotted_style(reference: str) -> ParsedReference:
    parts = reference.rsplit(".", 1)
    if len(parts) != _REFERENCE_PARTS_COUNT:
        msg = "点号形式引用 '{}' 非法;期望格式: `module.path.function`".format(reference)
        raise ReferenceSyntaxError(msg)

    module_path, func_name = parts
    if not module_path:
        msg = "点号形式引用 '{}' 非法;期望格式: `module.path.function`".format(reference)
        if reference.startswith("."):
            msg = "相对点号引用 '{}' 非法;期望格式: `.module.path.function`".format(reference)
        raise ReferenceSyntaxError(msg)

    if _MODULE_PATH_RE.fullmatch(module_path) is None:
        msg = "引用 '{}' 的模块路径 '{}' 非法".format(reference, module_path)
        raise ReferenceSyntaxError(msg)
    if not func_name.isidentifier():
        msg = "引用 '{}' 的可调用名 '{}' 非法".format(reference, func_name)
        raise ReferenceSyntaxError(msg)

    return ParsedReference(reference=reference, module_path=module_path, attr_path=(func_name,), style="dotted")


__all__ = [
    "REFERENCE_FORMAT_EXAMPLES",
    "ParsedReference",
    "ReferenceSyntaxError",
    "is_valid_python_reference",
    "parse_python_reference",
]



import contextlib
import time
from dataclasses import dataclass
from dataclasses import field as dataclass_field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

from ..hooks.base import HookManager, IExecutionHook
from ..ob.components import split_components
from ..ob.hub import InstrumentationHub
from ..ob.observability import Observability
from ..ob.observer import Observer
from ..ob.presets.viz import VizObserver, VizObserverConfig
from ..planning.builder import PlanBuilder
from ..planning.plan import ExecutionPlan
from ..sinks.sink_base import BaseRowSink, BaseSink, ColumnBatch, ColumnValues, IColumnSink, IRowSink, ISink
from ..sinks.sink_csv import ColumnCSVSink, CSVSink
from ..spec.ir.demand import DemandIr
from ..spec.ir.fields import DerivedFieldIr, FieldIr, SupportedFieldIr
from ..typedefs import ParallelMode, RowData, SinkRowKeySeq
from ..vendor.compact.typing_extensionsx import override
from .engine import ScalimEngine
from .guardrails import GuardrailsPolicy
from .loader_retry import LoaderRetryPolicies
from .output_contracts import ExportLayout, OutputSpec

if TYPE_CHECKING:
    from ..ob.manager import ObserverManager
    from .output_composition import OutputCompositionSpec, OutputTargetStats, RouterRowSink


@dataclass(frozen=True)
class ObservabilitySpec:
    """与 `DSL` 无关的可观测性请求(用于运行编排).

    - `viz_config` 可选;在构建执行计划后会物化为 `VizObserver`.
    """

    fallback_logger_enabled: bool = False
    viz_config: Optional[VizObserverConfig] = None


@dataclass(frozen=True)
class ExecutionRequest:
    export_layout: ExportLayout
    """导出布局(字段顺序与可选表头)."""

    output: OutputSpec = dataclass_field(default_factory=OutputSpec)
    """输出策略(例如输出格式、路径、编码、是否流式)."""

    sink: Optional[ISink] = None
    """可选:显式指定输出端;若为 `None` 则按 `output` 策略创建."""

    observability: Optional[ObservabilitySpec] = None
    """可选:可观测性请求(例如 `viz` 配置)."""

    components: Optional[List[Union[Observer, IExecutionHook]]] = None
    """可选:要挂载的 `Observer`/`Hook` 组件列表."""

    batch_size: Optional[int] = None
    """可选:覆盖批大小(`None` 表示不覆盖)."""

    parallel_mode: ParallelMode = "seq"
    """并行模式(`seq` 或 `adaptive`)."""

    max_workers: int = 0
    """最大并发工作数提示(`0` 表示自动)."""

    guardrails: Optional[GuardrailsPolicy] = None
    """可选:运行时护栏策略."""

    loader_retry: Optional[LoaderRetryPolicies] = None
    """可选:加载重试策略."""

    output_composition: Optional["OutputCompositionSpec"] = None
    """可选:多输出组合请求(`IR/Python-only`).

    当提供该字段时:
    - `output`/`sink` 的单输出装配将被忽略
    - 运行计划的目标字段将由组合请求的 `required_demand_fields` 计算得出
    """


@dataclass(frozen=True)
class ExecutionResult:
    """与 `DSL` 无关的执行结果.

    注意:
    - `total_rows` 统计写入到实际输出端的行数(包括 `NullSink`),这是输出/写出的行数.
    - 可观测性指标可能使用不同口径来做低开销吞吐估算
      (例如 `PerformanceMetrics.total_rows` 统计的是输入 `row_ids`).
    """

    output_path: Optional[str]
    total_rows: int
    duration: float
    demand_ir: DemandIr
    plan: ExecutionPlan
    outputs: Optional[Dict[str, str]] = None
    """可选:输出目标到 `output_path` 的映射(多输出组合时提供)."""

    output_target_stats: Optional[List["OutputTargetStats"]] = None
    """可选:每个输出目标的统计(行数/耗时/错误/禁用)(多输出组合时提供)."""


class _TeeRowSink(BaseRowSink):
    _primary: IRowSink
    _secondary: IRowSink

    def __init__(self, primary: IRowSink, secondary: IRowSink) -> None:
        self._primary = primary
        self._secondary = secondary

    @override
    def write_row(self, row: RowData) -> None:
        self._primary.write_row(row)
        self._secondary.write_row(row)

    @override
    def write_batch(self, rows: Sequence[RowData]) -> None:
        self._primary.write_batch(rows)
        self._secondary.write_batch(rows)

    @override
    def close(self) -> None:
        self._primary.close()
        self._secondary.close()


@dataclass(frozen=True)
class _OutputPlan:
    sink: ISink
    output_path: Optional[str]


class _TeeColumnSink(IColumnSink):
    _primary: IColumnSink
    _secondary: IColumnSink

    def __init__(self, primary: IColumnSink, secondary: IColumnSink) -> None:
        self._primary = primary
        self._secondary = secondary

    @override
    def write_batch(self, rows: Sequence[RowData]) -> None:
        self._primary.write_batch(rows)
        self._secondary.write_batch(rows)

    @override
    def set_row_ids(self, row_ids: "SinkRowKeySeq") -> None:
        self._primary.set_row_ids(row_ids)
        self._secondary.set_row_ids(row_ids)

    @override
    def write_column(self, field_key: str, values: ColumnValues) -> None:
        self._primary.write_column(field_key, values)
        self._secondary.write_column(field_key, values)

    @override
    def write_columns(self, columns: ColumnBatch) -> None:
        self._primary.write_columns(columns)
        self._secondary.write_columns(columns)

    @override
    def close(self) -> None:
        self._primary.close()
        self._secondary.close()


class _NullSink(BaseSink):
    @override
    def write_batch(self, rows: Sequence[RowData]) -> None:
        _ = rows

    @override
    def close(self) -> None:
        return


class InternalStatsCollector:
    """内部执行统计收集器.

    此对象刻意保持轻量,且与 `DSL` 无关.
    """

    def __init__(self) -> None:
        self.total_rows: int = 0


class _CountingRowSink(BaseRowSink):
    _sink: IRowSink
    _tracker: InternalStatsCollector

    def __init__(self, sink: IRowSink, tracker: InternalStatsCollector) -> None:
        self._sink = sink
        self._tracker = tracker

    @override
    def write_row(self, row: RowData) -> None:
        self._tracker.total_rows += 1
        self._sink.write_row(row)

    @override
    def write_batch(self, rows: Sequence[RowData]) -> None:
        self._tracker.total_rows += len(rows)
        self._sink.write_batch(rows)

    @override
    def close(self) -> None:
        self._sink.close()


class _CountingColumnSink(IColumnSink):
    _sink: IColumnSink
    _tracker: InternalStatsCollector

    def __init__(self, sink: IColumnSink, tracker: InternalStatsCollector) -> None:
        self._sink = sink
        self._tracker = tracker

    @override
    def set_row_ids(self, row_ids: "SinkRowKeySeq") -> None:
        self._tracker.total_rows += len(row_ids)
        self._sink.set_row_ids(row_ids)

    @override
    def write_column(self, field_key: str, values: ColumnValues) -> None:
        self._sink.write_column(field_key, values)

    @override
    def write_columns(self, columns: ColumnBatch) -> None:
        self._sink.write_columns(columns)

    @override
    def write_batch(self, rows: Sequence[RowData]) -> None:
        self._tracker.total_rows += len(rows)
        self._sink.write_batch(rows)

    @override
    def close(self) -> None:
        self._sink.close()


class _CountingBatchSink(ISink):
    _sink: ISink
    _tracker: InternalStatsCollector

    def __init__(self, sink: ISink, tracker: InternalStatsCollector) -> None:
        self._sink = sink
        self._tracker = tracker

    @override
    def write_batch(self, rows: Sequence[RowData]) -> None:
        self._tracker.total_rows += len(rows)
        self._sink.write_batch(rows)

    @override
    def close(self) -> None:
        self._sink.close()


def _wrap_sink_for_row_count(sink: ISink, tracker: InternalStatsCollector) -> ISink:
    if isinstance(sink, IColumnSink):
        return _CountingColumnSink(sink, tracker)
    if isinstance(sink, IRowSink):
        return _CountingRowSink(sink, tracker)
    return _CountingBatchSink(sink, tracker)


def _get_field_name(field_id: str, field_ir: SupportedFieldIr) -> str:
    name = getattr(field_ir, "name", "") or ""
    if name and name != field_id:
        return name
    return field_id


def export_layout_from_demand_ir(
    demand_ir: DemandIr,
    field_ids: Sequence[str],
    *,
    header_fields_output_by: str = "field_id",
) -> ExportLayout:
    """供适配器根据生效 `IR` 构建 `ExportLayout` 的辅助函数."""
    normalized_ids = tuple(str(item) for item in field_ids)
    if header_fields_output_by != "name":
        return ExportLayout(field_ids=normalized_ids, header_names=None)

    name_map: Dict[str, str] = {}
    for fid in normalized_ids:
        field_ir = demand_ir.fields.get(fid)
        if field_ir is None:
            continue
        resolved = _get_field_name(fid, field_ir)
        if resolved != fid:
            name_map[fid] = resolved

    if not name_map:
        return ExportLayout(field_ids=normalized_ids, header_names=None)

    header_names: Tuple[str, ...] = tuple(name_map.get(fid, fid) for fid in normalized_ids)
    return ExportLayout(field_ids=normalized_ids, header_names=header_names)


def _create_file_sink(output: OutputSpec, layout: ExportLayout) -> Optional[ISink]:
    if not output.path:
        return None

    output_path = Path(output.path)
    if output_path.parent and not output_path.parent.exists():
        output_path.parent.mkdir(parents=True, exist_ok=True)

    fmt = (output.format or "csv").lower()
    field_names = list(layout.field_ids)
    header_names = list(layout.header_names) if layout.header_names is not None else field_names

    if fmt == "csv":
        if output.streaming:
            return CSVSink(
                output_path=str(output.path),
                encoding=output.encoding,
                field_names=field_names,
                header_names=header_names,
                include_header=output.include_header,
                flush_policy="every_n_rows",
            )
        return ColumnCSVSink(
            output_path=str(output.path),
            field_names=field_names,
            header_names=header_names,
            encoding=output.encoding,
            include_header=output.include_header,
        )

    if fmt == "excel":
        from ..sinks.sink_excel import ColumnExcelSink, ExcelSink  # noqa: PLC0415

        if output.streaming:
            return ExcelSink(
                output_path=str(output.path),
                field_names=field_names,
                header_names=header_names,
                sheet_name=str(output.sheet_name) if output.sheet_name else "Sheet1",
                include_header=output.include_header,
                allow_formulas=bool(output.excel_allow_formulas),
                write_lock=bool(output.write_lock),
            )
        return ColumnExcelSink(
            output_path=str(output.path),
            field_names=field_names,
            header_names=header_names,
            sheet_name=str(output.sheet_name) if output.sheet_name else "Sheet1",
            include_header=output.include_header,
            allow_formulas=bool(output.excel_allow_formulas),
            write_lock=bool(output.write_lock),
        )

    msg = "Unsupported output format: '{}'. Supported formats: excel, csv.".format(output.format)
    raise ValueError(msg)


def _create_tee_sink(primary: ISink, secondary: ISink) -> ISink:
    if isinstance(primary, IRowSink) and isinstance(secondary, IRowSink):
        return _TeeRowSink(primary, secondary)
    if isinstance(primary, IColumnSink) and isinstance(secondary, IColumnSink):
        return _TeeColumnSink(primary, secondary)
    msg = "Incompatible sinks for tee: {} vs {}".format(type(primary).__name__, type(secondary).__name__)
    raise ValueError(msg)


def _describe_sink_kind(sink: ISink) -> str:
    if isinstance(sink, IRowSink):
        return "IRowSink"
    if isinstance(sink, IColumnSink):
        return "IColumnSink"
    return "ISink"


def _create_output_plan(output: OutputSpec, layout: ExportLayout, sink: Optional[ISink]) -> _OutputPlan:
    file_sink = _create_file_sink(output, layout)
    output_path: Optional[str] = output.path or None

    if sink is None:
        if file_sink is not None:
            return _OutputPlan(sink=file_sink, output_path=output_path)
        return _OutputPlan(sink=_NullSink(), output_path=None)

    if file_sink is None:
        return _OutputPlan(sink=sink, output_path=None)

    try:
        tee_sink = _create_tee_sink(file_sink, sink)
    except ValueError as e:
        msg = (
            "Incompatible sinks for tee: file_sink={}({}) vs sink={}({}). "
            "Both sinks must be IRowSink or both must be IColumnSink. "
            "Hint: set output.streaming=true for a row file sink (CSV/Excel) when teeing with an IRowSink; "
            "or use a column sink such as InMemoryColumnSink when output.streaming=false. "
            "If you pass a custom sink, it must implement IRowSink or IColumnSink to be tee-compatible."
        ).format(
            type(file_sink).__name__,
            _describe_sink_kind(file_sink),
            type(sink).__name__,
            _describe_sink_kind(sink),
        )
        with contextlib.suppress(Exception):
            file_sink.close()
        raise ValueError(msg) from e

    return _OutputPlan(sink=tee_sink, output_path=output_path)


@dataclass(frozen=True)
class _OutputAssembly:
    counting_sink: ISink
    output_path: Optional[str]
    outputs: Optional[Dict[str, str]]
    composition_router: Optional["RouterRowSink"]


def _build_execution_plan(demand_ir: DemandIr, request: ExecutionRequest) -> ExecutionPlan:
    plan_targets: List[str] = list(request.export_layout.field_ids)
    if request.output_composition is not None:
        from .output_composition import required_demand_fields  # noqa: PLC0415

        plan_targets = list(required_demand_fields(request.output_composition))
    return PlanBuilder(demand_ir).build(targets=plan_targets)


def _build_observer_and_hook_managers(
    *,
    plan: ExecutionPlan,
    request: ExecutionRequest,
    event_meta_defaults: Optional[Dict[str, Any]] = None,
) -> Tuple["ObserverManager", HookManager]:
    fallback_logger_enabled = False
    viz_config: Optional[VizObserverConfig] = None
    if request.observability is not None:
        fallback_logger_enabled = request.observability.fallback_logger_enabled
        viz_config = request.observability.viz_config

    observer_manager = Observability(fallback_logger_enabled=fallback_logger_enabled).build_manager(event_meta_defaults=event_meta_defaults)

    component_observers, component_hooks = split_components(request.components)
    for observer in component_observers:
        observer_manager.register(observer)

    if viz_config is not None:
        observer_manager.register(VizObserver.from_plan(plan, viz_config, output_composition=request.output_composition))

    hook_manager = HookManager(fallback_logger_enabled=fallback_logger_enabled)
    for hook in component_hooks:
        hook_manager.register(hook)

    return observer_manager, hook_manager


def _build_field_fingerprints_for_meta(demand_ir: DemandIr) -> List[Tuple[str, str, str, str]]:
    """生成稳定指纹(不包含可调用对象`callable`),用于元信息工作表."""
    field_fingerprints: List[Tuple[str, str, str, str]] = []
    for field_id in sorted(demand_ir.fields.keys()):
        spec = demand_ir.fields[field_id]
        if isinstance(spec, FieldIr):
            field_fingerprints.append((spec.field_id, "field", str(spec.source.source_id), str(spec.data_key)))
        elif isinstance(spec, DerivedFieldIr):
            field_fingerprints.append((spec.field_id, "derived", "", ",".join(spec.dependencies)))
        else:
            field_fingerprints.append((str(field_id), type(spec).__name__, "", ""))
    return field_fingerprints


def _select_primary_output_path(outputs: Dict[str, str], spec: "OutputCompositionSpec") -> Optional[str]:
    primary_id: Optional[str] = None
    for t in spec.targets:
        if t.is_primary:
            primary_id = str(t.target_id)
            break
    if primary_id is None:
        for t in spec.derived_targets:
            if t.is_primary:
                primary_id = str(t.target_id)
                break

    if primary_id is not None:
        resolved = outputs.get(primary_id)
        if resolved is not None:
            return resolved

    if outputs:
        return next(iter(outputs.values()))
    return None


def _assemble_outputs(
    *,
    demand_ir: DemandIr,
    plan: ExecutionPlan,
    request: ExecutionRequest,
    hook_manager: HookManager,
    observer_manager: "ObserverManager",
    wall_start_time: float,
    batch_size: Optional[int],
    stats: InternalStatsCollector,
) -> _OutputAssembly:
    composition_spec = request.output_composition
    if composition_spec is None:
        output_plan = _create_output_plan(request.output, request.export_layout, request.sink)
        counting_sink = _wrap_sink_for_row_count(output_plan.sink, stats)
        return _OutputAssembly(
            counting_sink=counting_sink,
            output_path=output_plan.output_path,
            outputs=None,
            composition_router=None,
        )

    from .output_composition import build_output_composition  # noqa: PLC0415

    # 构建一个与 `engine` 共享订阅者的 `InstrumentationHub`,用于输出级事件(每个输出目标结束统计).
    instrumentation = InstrumentationHub(hook_manager=hook_manager, observer_manager=observer_manager)

    composition_plan = build_output_composition(
        spec=composition_spec,
        demand_name=str(demand_ir.name),
        demand_main_source_id=str(demand_ir.main_source.source_id),
        demand_target_fields=list(plan.target_fields),
        demand_field_fingerprints=_build_field_fingerprints_for_meta(demand_ir),
        run_started_at_epoch=wall_start_time,
        run_parallel_mode=str(request.parallel_mode),
        run_batch_size=batch_size,
        instrumentation=instrumentation,
    )

    router_sink = composition_plan.sink
    composed_sink: ISink = router_sink
    if request.sink is not None:
        try:
            composed_sink = _create_tee_sink(router_sink, request.sink)
        except ValueError as e:
            msg = (
                "Incompatible sinks for tee: composed_sink={}({}) vs sink={}({}). "
                "Both sinks must be IRowSink (output_composition only supports streaming row sinks). "
                "Hint: use InMemoryRowSink (or another IRowSink) when teeing with composed outputs."
            ).format(
                type(router_sink).__name__,
                _describe_sink_kind(router_sink),
                type(request.sink).__name__,
                _describe_sink_kind(request.sink),
            )
            with contextlib.suppress(Exception):
                router_sink.close()
            raise ValueError(msg) from e

    counting_sink = _wrap_sink_for_row_count(composed_sink, stats)
    outputs = composition_plan.output_paths
    output_path = _select_primary_output_path(outputs, composition_spec) if outputs else None
    return _OutputAssembly(
        counting_sink=counting_sink,
        output_path=output_path,
        outputs=outputs,
        composition_router=router_sink,
    )


def _create_engine_with_cleanup(
    *,
    demand_ir: DemandIr,
    plan: ExecutionPlan,
    request: ExecutionRequest,
    hook_manager: HookManager,
    observer_manager: "ObserverManager",
    batch_size: Optional[int],
    sink: ISink,
    engine_factory: Optional[Callable[..., ScalimEngine]] = None,
) -> ScalimEngine:
    engine_cls = engine_factory or ScalimEngine
    try:
        return engine_cls(
            demand=demand_ir,
            plan=plan,
            hook_manager=hook_manager,
            observer_manager=observer_manager,
            guardrails=request.guardrails,
            loader_retry=request.loader_retry,
            batch_size=batch_size,
            parallel_mode=request.parallel_mode,
            max_workers=request.max_workers,
        )
    except Exception:
        with contextlib.suppress(Exception):
            sink.close()
        with contextlib.suppress(Exception):
            observer_manager.close()
        raise


def run_ir(
    demand_ir: DemandIr,
    request: ExecutionRequest,
    engine_factory: Optional[Callable[..., ScalimEngine]] = None,
    event_meta_defaults: Optional[Dict[str, Any]] = None,
) -> ExecutionResult:
    start_time = time.perf_counter()
    wall_start_time = time.time()

    plan = _build_execution_plan(demand_ir, request)
    observer_manager, hook_manager = _build_observer_and_hook_managers(plan=plan, request=request, event_meta_defaults=event_meta_defaults)

    stats = InternalStatsCollector()
    batch_size = request.batch_size if request.batch_size is not None else demand_ir.batch_size_hint

    output_assembly = _assemble_outputs(
        demand_ir=demand_ir,
        plan=plan,
        request=request,
        hook_manager=hook_manager,
        observer_manager=observer_manager,
        wall_start_time=wall_start_time,
        batch_size=batch_size,
        stats=stats,
    )

    engine = _create_engine_with_cleanup(
        demand_ir=demand_ir,
        plan=plan,
        request=request,
        hook_manager=hook_manager,
        observer_manager=observer_manager,
        batch_size=batch_size,
        sink=output_assembly.counting_sink,
        engine_factory=engine_factory,
    )

    try:
        _ = engine.run(sink=output_assembly.counting_sink)
    finally:
        # 尽力清理: 即使 `engine`/`pipeline` 在关闭前失败也要执行.
        with contextlib.suppress(Exception):
            output_assembly.counting_sink.close()
        with contextlib.suppress(Exception):
            observer_manager.close()

    output_target_stats: Optional[List["OutputTargetStats"]] = None
    if output_assembly.composition_router is not None:
        # 注意: `counting_sink.close()` 在 `finally` 中执行;此处读取的是关闭后的最终统计快照.
        output_target_stats = output_assembly.composition_router.get_target_stats()

    return ExecutionResult(
        output_path=output_assembly.output_path,
        total_rows=stats.total_rows,
        duration=time.perf_counter() - start_time,
        demand_ir=demand_ir,
        plan=plan,
        outputs=output_assembly.outputs,
        output_target_stats=output_target_stats,
    )


__all__ = [
    "ExecutionRequest",
    "ExecutionResult",
    "ExportLayout",
    "ObservabilitySpec",
    "OutputSpec",
    "export_layout_from_demand_ir",
    "run_ir",
]



import threading
from collections.abc import MutableMapping as MutableMappingABC
from typing import TYPE_CHECKING, Callable, Dict, Iterator

from ..typedefs import LoaderResultMapping
from ..vendor.compact.typing_extensionsx import override

if TYPE_CHECKING:
    from typing import MutableMapping as MutableMappingType

    _PreloadCacheBase = MutableMappingType[str, LoaderResultMapping]
else:
    _PreloadCacheBase = MutableMappingABC


class PreloadCache(_PreloadCacheBase):
    """线程安全的 `preload_forever` 缓存容器(按 `source_id` 分桶去重).

    设计目标:
    - 支持跨多个 `runs` 共享 `preload_forever` 结果(`Workflow` 场景)
    - 并发下对单个 `source_id` 加锁,保证最多一次真实 `loader` 调用
    """

    def __init__(self) -> None:
        self._data: Dict[str, LoaderResultMapping] = {}
        self._locks: Dict[str, threading.Lock] = {}
        self._global_lock: threading.Lock = threading.Lock()

    def __getstate__(self) -> Dict[str, object]:
        # 仅序列化数据,锁在反序列化后重建.
        # 说明:
        # - `adaptive` 的 `process` 后端会 `pickle` `preloaded_cache`,因此此对象必须可被 `pickle`.
        # - 共享锁只在同一进程内有意义.
        return {"_data": self._data}

    def __setstate__(self, state: Dict[str, object]) -> None:
        raw_data = state.get("_data", {})
        if not isinstance(raw_data, dict):
            raw_data = {}
        self._data = raw_data  # type: ignore[assignment]
        self._locks = {}
        self._global_lock = threading.Lock()

    @override
    def __getitem__(self, key: str) -> LoaderResultMapping:
        return self._data[key]

    @override
    def __setitem__(self, key: str, value: LoaderResultMapping) -> None:
        self._data[key] = value

    @override
    def __delitem__(self, key: str) -> None:
        del self._data[key]

    @override
    def __iter__(self) -> Iterator[str]:
        return iter(self._data)

    @override
    def __len__(self) -> int:
        return len(self._data)

    def get_or_load(self, source_id: str, load_fn: Callable[[], LoaderResultMapping]) -> LoaderResultMapping:
        existing = self._data.get(source_id)
        if existing is not None:
            return existing

        lock = self._lock_for(source_id)
        with lock:
            existing = self._data.get(source_id)
            if existing is not None:
                return existing
            value = load_fn()
            self._data[source_id] = value
            return value

    def _lock_for(self, source_id: str) -> threading.Lock:
        with self._global_lock:
            lock = self._locks.get(source_id)
            if lock is None:
                lock = threading.Lock()
                self._locks[source_id] = lock
            return lock


__all__ = [
    "PreloadCache",
]



import heapq
import logging
from typing import Dict, List, Set, Tuple, Union

from ...spec.ir.sources import SourceIr
from ...utils import graph

_logger = logging.getLogger(__name__)
_UNMAPPED_DEPS_PREVIEW_LIMIT = 10

REF_LOADER_ORDERING_DEGRADED_PREFIX = "引用加载器排序退化"
REF_LOADER_ORDERING_DEGRADED_WARNING = (
    REF_LOADER_ORDERING_DEGRADED_PREFIX + ": 有 %d 个依赖键无法映射到对应的引用加载器; 将回退到稳定的并列裁决(`source_id`). 键=%s"
)


def sort_ref_loaders(  # noqa: C901, PLR0912
    ref_loaders: List[Tuple[SourceIr, List[Tuple[str, Union[str, Tuple[str, ...]]]]]],
) -> List[Tuple[SourceIr, List[Tuple[str, Union[str, Tuple[str, ...]]]]]]:
    """基于引用字段的依赖信号对引用加载器进行拓扑排序.

    排序约束:
    - 依赖信号来自引用字段的关联步骤(`from_field` 对应的字段键).
    - 排序器会将这些字段键映射回其所属的引用加载器,构建加载器依赖图并执行拓扑排序.
    - 如果部分依赖键无法映射,排序器会仅警告一次,并回退到稳定的并列裁决(按 `source_id` 字典序).
    """
    if len(ref_loaders) <= 1:
        return ref_loaders

    loader_ids = [source.source_id for source, _ in ref_loaders]
    loader_deps: Dict[str, Set[str]] = {loader_id: set() for loader_id in loader_ids}
    field_to_loader: Dict[str, str] = {}

    for source, ref_fields in ref_loaders:
        for field_key, _ in ref_fields:
            field_to_loader[field_key] = source.source_id

    unmapped_dep_keys: Set[str] = set()
    for source, ref_fields in ref_loaders:
        loader_id = source.source_id
        for _field_key, dep_ref_field_keys in ref_fields:
            dep_keys = [dep_ref_field_keys] if isinstance(dep_ref_field_keys, str) else list(dep_ref_field_keys)
            for dep_key in dep_keys:
                if not dep_key:
                    continue
                dep_loader = field_to_loader.get(dep_key)
                if dep_loader is None:
                    unmapped_dep_keys.add(dep_key)
                    continue
                if dep_loader != loader_id:
                    loader_deps[loader_id].add(dep_loader)

    if unmapped_dep_keys:
        unmapped_preview = ", ".join(sorted(unmapped_dep_keys)[:_UNMAPPED_DEPS_PREVIEW_LIMIT])
        if len(unmapped_dep_keys) > _UNMAPPED_DEPS_PREVIEW_LIMIT:
            unmapped_preview += ", ..."
        _logger.warning(REF_LOADER_ORDERING_DEGRADED_WARNING, len(unmapped_dep_keys), unmapped_preview)

    in_degree: Dict[str, int] = dict.fromkeys(loader_ids, 0)
    reverse_deps: Dict[str, List[str]] = {loader_id: [] for loader_id in loader_ids}
    for loader_id, deps in loader_deps.items():
        for dep in deps:
            in_degree[loader_id] += 1
            reverse_deps.setdefault(dep, []).append(loader_id)

    ready = [name for name, degree in in_degree.items() if degree == 0]
    heapq.heapify(ready)
    sorted_loaders: List[str] = []
    while ready:
        name = heapq.heappop(ready)
        sorted_loaders.append(name)
        for dependent in sorted(reverse_deps.get(name, [])):
            in_degree[dependent] -= 1
            if in_degree[dependent] == 0:
                heapq.heappush(ready, dependent)

    if len(sorted_loaders) != len(in_degree):
        cycles = graph.detect_cycles(loader_ids, lambda name: sorted(loader_deps.get(name, set())))
        cycle_str = "; ".join(" -> ".join(cycle) for cycle in cycles) if cycles else "<unknown>"
        msg = "检测到 ref loader 循环依赖: {}".format(cycle_str)
        raise graph.CyclicDependencyError(msg, cycles)

    loader_map = {source.source_id: (source, fields) for source, fields in ref_loaders}
    return [loader_map[name] for name in sorted_loaders if name in loader_map]


__all__ = [
    "sort_ref_loaders",
]



import copy
import json
from dataclasses import Field, dataclass
from dataclasses import fields as dataclass_fields
from pathlib import Path
from types import SimpleNamespace
from typing import Any, ClassVar, Dict, List, Optional, Set, Tuple, Type, Union, cast

from . import constants as schema_constants
from . import models as schema_models
from .constants import SCHEMA_META_KEY

_IMPORT_KEY = "$import"
_IMPORTS_KEY = "imports"


def _build_default_types_module() -> Any:
    merged: Dict[str, Any] = {}
    # 仅复制公开名称.这样可以保持默认模块显式化,并避免动态属性回退.
    for name, value in vars(schema_constants).items():
        if name.startswith("_"):
            continue
        merged[name] = value
    for name, value in vars(schema_models).items():
        if name.startswith("_"):
            continue
        merged[name] = value
    return SimpleNamespace(**merged)


_DEFAULT_TYPES_MODULE = _build_default_types_module()


@dataclass(frozen=True)
class SchemaMeta:
    schema_name: Optional[str]
    meta: Dict[str, Any]

    @classmethod
    def from_field(cls, dc_field: "Field[Any]") -> "SchemaMeta":
        raw_meta = dict(dc_field.metadata.get(SCHEMA_META_KEY, {}))
        schema_name = raw_meta.pop("schema_name", None)
        if not isinstance(schema_name, str):
            schema_name = None
        return cls(schema_name=schema_name, meta=raw_meta)


class SchemaBuilder:
    META_KEY_MAP: ClassVar[Dict[str, str]] = {
        "desc": "description",
        "md": "markdownDescription",
        "markdown": "markdownDescription",
        "choices": "enum",
        "min": "minimum",
        "max": "maximum",
        "min_items": "minItems",
        "max_items": "maxItems",
        "min_props": "minProperties",
        "max_props": "maxProperties",
        "pattern": "pattern",
        "default": "default",
        "type": "type",
        "items": "items",
        "one_of": "oneOf",
        "any_of": "anyOf",
        "all_of": "allOf",
        "additional_props": "additionalProperties",
        "const": "const",
        "deprecated": "deprecated",
        "items_choices": "items_choices",
        "example": "examples",
    }
    ORDER_INSENSITIVE_KEYS: ClassVar[Set[str]] = {"required", "enum", "oneOf", "anyOf", "allOf"}
    IGNORED_KEYS: ClassVar[Set[str]] = {"$comment"}
    ELLIPSIS_TUPLE_LEN: ClassVar[int] = 2
    GENERATED_SCHEMA_COMMENT: ClassVar[str] = "自动生成, 请勿手动修改. 生成脚本: scripts/gen-yaml-dsl-schema.py"
    PRIMITIVE_TYPE_MAP: ClassVar[Dict[Type[Any], str]] = {
        bool: "boolean",
        int: "integer",
        str: "string",
    }
    _types: Any

    def __init__(self, types_module: Optional[Any] = None) -> None:
        if types_module is None:
            types_module = _DEFAULT_TYPES_MODULE
        self._types = types_module

    def build_demand_schema(self) -> Dict[str, Any]:
        types_mod = self._types
        definitions = {
            "main_source": self._build_definition(types_mod.MainSourceConfig),
            "source": self._build_definition(types_mod.SourceConfig),
            "source_field_inline": self._build_definition(types_mod.SourceFieldConfig),
            "field": self._build_field_definition(),
            "relation": self._build_definition(types_mod.RelationConfig),
            "output_container": self._build_definition(types_mod.OutputContainerConfig),
            "output_aggregate": self._build_definition(types_mod.OutputAggregateConfig),
            "output_target": self._build_definition(types_mod.OutputTargetConfig),
            "output_extra_sheet": self._build_definition(types_mod.OutputExtraSheetConfig),
            "logging": self._build_definition(types_mod.LoggingConfig),
            "performance_thresholds": self._build_definition(types_mod.PerformanceThresholdsConfig),
            "performance_report": self._build_definition(types_mod.PerformanceReportConfig),
            "performance": self._build_definition(types_mod.PerformanceConfig),
            "relation_report": self._build_definition(types_mod.RelationReportConfig),
            "relations": self._build_definition(types_mod.RelationsConfig),
            "viz": self._build_definition(types_mod.VizConfig),
            "trace": self._build_definition(types_mod.TraceConfig),
            "row_gap": self._build_definition(types_mod.RowGapConfig),
            "memory_opt": self._build_definition(types_mod.MemoryOptimizationConfig),
            "observability": self._build_definition(types_mod.ObservabilityConfig),
            "guardrails_loader": self._build_definition(types_mod.GuardrailsLoaderConfig),
            "guardrails_relations": self._build_definition(types_mod.GuardrailsRelationsConfig),
            "guardrails_compute": self._build_definition(types_mod.GuardrailsComputeConfig),
            "guardrails": self._build_definition(types_mod.GuardrailsConfig),
            "loader_retry": self._build_definition(types_mod.LoaderRetryConfig),
        }

        schema: Dict[str, Any] = {
            "$schema": types_mod.DEMAND_SCHEMA_META["$schema"],
            "$id": types_mod.DEMAND_SCHEMA_META["$id"],
            "title": types_mod.DEMAND_SCHEMA_META["title"],
            "description": types_mod.DEMAND_SCHEMA_META["description"],
            "$comment": self.GENERATED_SCHEMA_COMMENT,
            "type": "object",
            # `$import` 会在编译期展开;为提升 `LSP`/`schema` 体验,允许仅声明 `$import` 的用法通过校验.
            "anyOf": [{"required": list(types_mod.DEMAND_SCHEMA_REQUIRED)}, {"required": [_IMPORT_KEY]}],
            "properties": self._build_demand_properties(),
            "definitions": definitions,
        }
        if "markdownDescription" in types_mod.DEMAND_SCHEMA_META:
            schema["markdownDescription"] = types_mod.DEMAND_SCHEMA_META["markdownDescription"]
        additional_props = getattr(types_mod.DemandConfig, "SCHEMA_ADDITIONAL_PROPERTIES", None)
        if additional_props is not None:
            schema["additionalProperties"] = bool(additional_props)
        return schema

    def build_workflow_schema(self) -> Dict[str, Any]:
        types_mod = self._types
        cache_pool_pin_item: Dict[str, Any] = {
            "type": "object",
            "required": ["kind", "source_id"],
            "properties": {
                "kind": {
                    "type": "string",
                    "enum": ["preload_forever"],
                    "description": "pin kind(v0 仅允许 preload_forever)",
                    "markdownDescription": "pin kind(v0 仅允许 `preload_forever`).",
                },
                "source_id": {
                    "type": "string",
                    "minLength": 1,
                    "description": "pin 的 source_id",
                    "markdownDescription": "pin 的 `source_id`.",
                },
            },
            "additionalProperties": False,
        }

        cache_pool: Dict[str, Any] = {
            "type": "object",
            "required": ["conflict_policy", "release_policy", "budget"],
            "properties": {
                "conflict_policy": {
                    "type": "string",
                    "enum": ["error", "separate", "warn"],
                    "description": "signature 冲突策略(error/separate/warn)",
                    "markdownDescription": "signature 冲突策略(`error`/`separate`/`warn`).",
                },
                "release_policy": {
                    "type": "string",
                    "enum": ["dag_refcount", "workflow_end"],
                    "description": "释放策略(dag_refcount/workflow_end)",
                    "markdownDescription": "释放策略(`dag_refcount`/`workflow_end`).",
                },
                "budget": {
                    "type": "object",
                    "required": ["max_entries", "over_budget_policy"],
                    "properties": {
                        "max_entries": {
                            "type": "integer",
                            "minimum": 1,
                            "description": "cache pool entry 数量预算(>=1)",
                            "markdownDescription": "cache pool entry 数量预算(>=1).",
                        },
                        "over_budget_policy": {
                            "type": "string",
                            "enum": ["fail_fast", "evict_lru"],
                            "description": "超限策略(fail_fast/evict_lru)",
                            "markdownDescription": "超限策略(`fail_fast`/`evict_lru`).",
                        },
                    },
                    "additionalProperties": False,
                },
                "pin": {
                    "type": "array",
                    "items": cache_pool_pin_item,
                    "default": [],
                    "description": "pin 列表(可选)",
                    "markdownDescription": "pin 列表(可选).",
                },
            },
            "additionalProperties": False,
        }

        options: Dict[str, Any] = {
            "type": "object",
            "properties": {
                "max_concurrency": {
                    "type": "integer",
                    "minimum": 1,
                    "default": 1,
                    "description": "runs 粒度并发上限(>=1)",
                    "markdownDescription": "runs 粒度并发上限(>=1).",
                },
                "failure_policy": {
                    "type": "string",
                    "enum": ["all_fail", "primary_only"],
                    "default": "all_fail",
                    "description": "失败策略(all_fail/primary_only)",
                    "markdownDescription": (
                        "失败策略.\n\n- `all_fail`: 任一 run 失败即失败\n- `primary_only`: 失败 run 被跳过但 workflow 继续"
                    ),
                },
                "cache_pool": {
                    "oneOf": [cache_pool, {"type": "null"}],
                    "default": None,
                    "description": "workflow-scope cache pool 配置(可选)",
                    "markdownDescription": "workflow-scope cache pool 配置(可选).",
                },
                "ctx": {
                    "oneOf": [
                        {
                            "type": "object",
                            "properties": {
                                "max_value_bytes": {
                                    "type": "integer",
                                    "minimum": 1,
                                    "default": 65536,
                                    "description": "ctx 单 key 最大字节数(>=1)",
                                    "markdownDescription": "ctx 单 key 最大字节数(>=1).",
                                },
                                "max_bytes": {
                                    "type": "integer",
                                    "minimum": 1,
                                    "default": 1048576,
                                    "description": "ctx 总量最大字节数(>=1)",
                                    "markdownDescription": "ctx 总量最大字节数(>=1).",
                                },
                            },
                            "additionalProperties": False,
                        },
                        {"type": "null"},
                    ],
                    "default": None,
                    "description": "workflow-level ctx 护栏配置(可选)",
                    "markdownDescription": "workflow-level ctx 护栏配置(可选).",
                },
            },
            "additionalProperties": False,
        }

        workbook_resource: Dict[str, Any] = {
            "type": "object",
            "required": ["path"],
            "properties": {
                "path": {
                    "type": "string",
                    "minLength": 1,
                    "description": "workbook 输出路径(非空字符串)",
                    "markdownDescription": "workbook 输出路径(非空字符串).",
                    "examples": ["./out/report.xlsx"],
                }
            },
            "additionalProperties": False,
        }

        csv_resource: Dict[str, Any] = {
            "type": "object",
            "required": ["path"],
            "properties": {
                "path": {
                    "type": "string",
                    "minLength": 1,
                    "description": "csv 输出路径(非空字符串)",
                    "markdownDescription": "csv 输出路径(非空字符串).",
                    "examples": ["./out/report.csv"],
                }
            },
            "additionalProperties": False,
        }

        write_to_workbook_sheet: Dict[str, Any] = {
            "type": "object",
            "required": ["workbook", "sheet", "output"],
            "properties": {
                "workbook": {"type": "string", "minLength": 1},
                "sheet": {"type": "string", "minLength": 1},
                "output": {"type": "string", "minLength": 1},
                "on_conflict": {
                    "type": "string",
                    "enum": ["error", "overwrite", "skip"],
                    "default": "error",
                },
            },
            "additionalProperties": False,
        }

        write_to_workbook_append: Dict[str, Any] = {
            "type": "object",
            "required": ["workbook", "sheet", "output"],
            "properties": {
                "workbook": {"type": "string", "minLength": 1},
                "sheet": {"type": "string", "minLength": 1},
                "output": {"type": "string", "minLength": 1},
                "align_by": {
                    "type": "string",
                    "enum": ["field_id", "header"],
                    "default": "field_id",
                },
                "header_policy": {
                    "type": "string",
                    "enum": ["once", "always", "never"],
                    "default": "once",
                },
                "on_mismatch": {
                    "type": "string",
                    "enum": ["error", "warn", "skip"],
                    "default": "error",
                },
            },
            "additionalProperties": False,
        }

        write_to_csv_append: Dict[str, Any] = {
            "type": "object",
            "required": ["csv", "output"],
            "properties": {
                "csv": {"type": "string", "minLength": 1},
                "output": {"type": "string", "minLength": 1},
                "header_policy": {
                    "type": "string",
                    "enum": ["once", "always", "never"],
                    "default": "once",
                },
                "on_mismatch": {
                    "type": "string",
                    "enum": ["error", "warn", "skip"],
                    "default": "error",
                },
            },
            "additionalProperties": False,
        }

        sheetbook_budget: Dict[str, Any] = {
            "type": "object",
            "required": ["max_sheets", "max_total_cells"],
            "properties": {
                "max_sheets": {
                    "type": "integer",
                    "minimum": 1,
                    "description": "sheetbook budget: max_sheets(正整数)",
                    "markdownDescription": "sheetbook budget: `max_sheets`(正整数).",
                    "examples": [32],
                },
                "max_total_cells": {
                    "type": "integer",
                    "minimum": 1,
                    "description": "sheetbook budget: max_total_cells(正整数)",
                    "markdownDescription": "sheetbook budget: `max_total_cells`(正整数).",
                    "examples": [5000000],
                },
            },
            "additionalProperties": False,
        }

        sheetbook_export_xlsx: Dict[str, Any] = {
            "type": "object",
            "required": ["path"],
            "properties": {
                "path": {
                    "type": "string",
                    "minLength": 1,
                    "description": "导出 xlsx 路径(非空字符串)",
                    "markdownDescription": "导出 xlsx 路径(非空字符串).",
                    "examples": ["./out/report.xlsx"],
                },
                "write_lock": {
                    "type": "boolean",
                    "default": False,
                    "description": "导出阶段启用写锁(可选)",
                    "markdownDescription": "导出阶段启用写锁(可选).",
                    "examples": [True],
                },
            },
            "additionalProperties": False,
        }

        sheetbook_resource: Dict[str, Any] = {
            "type": "object",
            "required": ["budget"],
            "properties": {
                "budget": sheetbook_budget,
                "export_xlsx": sheetbook_export_xlsx,
            },
            "additionalProperties": False,
        }

        write_to_sheetbook_sheet: Dict[str, Any] = {
            "type": "object",
            "required": ["sheetbook", "sheet", "output"],
            "properties": {
                "sheetbook": {"type": "string", "minLength": 1},
                "sheet": {"type": "string", "minLength": 1},
                "output": {"type": "string", "minLength": 1},
                "on_conflict": {
                    "type": "string",
                    "enum": ["error", "overwrite", "skip"],
                    "default": "error",
                },
            },
            "additionalProperties": False,
        }

        write_to_sheetbook_append: Dict[str, Any] = {
            "type": "object",
            "required": ["sheetbook", "sheet", "output"],
            "properties": {
                "sheetbook": {"type": "string", "minLength": 1},
                "sheet": {"type": "string", "minLength": 1},
                "output": {"type": "string", "minLength": 1},
                "align_by": {
                    "type": "string",
                    "enum": ["field_id", "header"],
                    "default": "field_id",
                },
                "header_policy": {
                    "type": "string",
                    "enum": ["once", "always", "never"],
                    "default": "once",
                },
                "on_mismatch": {
                    "type": "string",
                    "enum": ["error", "warn", "skip"],
                    "default": "error",
                },
            },
            "additionalProperties": False,
        }

        write_to: Dict[str, Any] = {
            "oneOf": [
                {"type": "null"},
                {
                    "type": "object",
                    "required": ["workbook_sheet"],
                    "properties": {"workbook_sheet": write_to_workbook_sheet},
                    "additionalProperties": False,
                },
                {
                    "type": "object",
                    "required": ["workbook_append"],
                    "properties": {"workbook_append": write_to_workbook_append},
                    "additionalProperties": False,
                },
                {
                    "type": "object",
                    "required": ["csv_append"],
                    "properties": {"csv_append": write_to_csv_append},
                    "additionalProperties": False,
                },
                {
                    "type": "object",
                    "required": ["sheetbook_sheet"],
                    "properties": {"sheetbook_sheet": write_to_sheetbook_sheet},
                    "additionalProperties": False,
                },
                {
                    "type": "object",
                    "required": ["sheetbook_append"],
                    "properties": {"sheetbook_append": write_to_sheetbook_append},
                    "additionalProperties": False,
                },
            ],
            "default": None,
            "description": "共享输出写入意图简写(可选)",
            "markdownDescription": "共享输出写入意图简写(可选).\n\n- MUST 恰好选择一个 write intent",
        }

        run_item: Dict[str, Any] = {
            "type": "object",
            "required": ["id", "demand"],
            "properties": {
                "id": {
                    "type": "string",
                    "minLength": 1,
                    "description": "run 标识(非空且唯一)",
                    "markdownDescription": "run 标识(非空且唯一).",
                },
                "demand": {
                    "type": "string",
                    "minLength": 1,
                    "description": "demand YAML 路径(字符串)",
                    "markdownDescription": (
                        "demand YAML 路径(字符串).\n\n"
                        "- 相对路径以 workflow 文件所在目录为基准\n"
                        "- 可通过 Python 入口注入 path_aliases 解析 `@/...` 或 `ALIAS:/...`"
                    ),
                },
                "depends_on": {
                    "type": "array",
                    "items": {"type": "string", "minLength": 1},
                    "default": [],
                    "description": "显式依赖 run.id 列表(可选)",
                    "markdownDescription": "显式依赖 `run.id` 列表(可选).",
                },
                "init_vars": {
                    "oneOf": [
                        {
                            "type": "object",
                            "propertyNames": {"type": "string", "minLength": 1},
                            "additionalProperties": True,
                        },
                        {"type": "null"},
                    ],
                    "default": None,
                    "description": "demand compile-time init_vars(可选,支持 $ctx 指令)",
                    "markdownDescription": "demand compile-time `init_vars`(可选,支持 `$ctx` 指令).",
                },
                "write_to": write_to,
            },
            "additionalProperties": False,
        }

        workflow: Dict[str, Any] = {
            "type": "object",
            "required": ["runs"],
            "properties": {
                "runs": {
                    "type": "array",
                    "minItems": 1,
                    "items": run_item,
                },
                "options": options,
                "resources": {
                    "type": "object",
                    "properties": {
                        "workbooks": {
                            "type": "object",
                            "default": {},
                            "propertyNames": {"type": "string", "minLength": 1},
                            "additionalProperties": workbook_resource,
                        },
                        "csvs": {
                            "type": "object",
                            "default": {},
                            "propertyNames": {"type": "string", "minLength": 1},
                            "additionalProperties": csv_resource,
                        },
                        "sheetbooks": {
                            "type": "object",
                            "default": {},
                            "propertyNames": {"type": "string", "minLength": 1},
                            "additionalProperties": sheetbook_resource,
                        },
                    },
                    "description": "workflow-scope shared output resources",
                    "markdownDescription": "workflow-scope shared output resources.",
                    "additionalProperties": False,
                },
            },
            "additionalProperties": False,
        }

        schema: Dict[str, Any] = {
            "$schema": types_mod.DEMAND_SCHEMA_META["$schema"],
            "$id": "https://scalim.example.com/schemas/workflow.json",
            "title": "Scalim Workflow 配置",
            "description": "Scalim 框架 workflow YAML 配置定义 Schema",
            "$comment": self.GENERATED_SCHEMA_COMMENT,
            "type": "object",
            "required": ["workflow"],
            "properties": {"workflow": workflow},
            "additionalProperties": False,
        }
        return schema

    def _import_ref_schema(self) -> Dict[str, Any]:
        return {
            "oneOf": [
                {"type": "string"},
                {"type": "array", "items": {"type": "string"}, "minItems": 1},
            ],
            "description": "$import 引用(支持 string 或 string list)",
            "markdownDescription": (
                "$import 引用.\n\n"
                "- string: `<alias>(.<segment>)*`\n"
                "- list: 按顺序合并,后者覆盖前者,最终再被本地覆盖\n"
                "- 仅支持 mapping 片段\n"
                "- V1 仅支持同级文件导入(见顶层 `imports`)"
            ),
            "examples": ["common.sources", ["common.sources", "other.sources"]],
        }

    def _imports_schema(self) -> Dict[str, Any]:
        return {
            "type": "object",
            "description": "片段文件导入别名映射(编译期展开)",
            "markdownDescription": (
                "片段文件导入别名映射.\n\n"
                "- key: alias\n"
                "- value: 片段文件路径(字符串)\n"
                "- V1 仅支持同级文件名: `x.yaml|x.yml` 或 `./x.yaml|./x.yml`\n"
                "- 禁止: 绝对路径/父目录/子目录/alias 前缀"
            ),
            "propertyNames": {"type": "string", "pattern": r"^[a-zA-Z_][a-zA-Z0-9_]*$"},
            "additionalProperties": {
                "type": "string",
                "pattern": r"^(\./)?[^/\\:]+\.ya?ml$",
            },
        }

    def _build_definition(self, cls: type) -> Dict[str, Any]:
        properties = self._build_class_properties(cls)
        properties.setdefault(_IMPORT_KEY, self._import_ref_schema())
        schema: Dict[str, Any] = {
            "type": "object",
            "properties": properties,
        }

        required = getattr(cls, "SCHEMA_REQUIRED", ())
        if required:
            # `$import` 会在编译期展开;为提升 `LSP`/`schema` 体验,允许仅声明 `$import` 的用法通过校验.
            schema["anyOf"] = [{"required": list(required)}, {"required": [_IMPORT_KEY]}]

        additional_props = getattr(cls, "SCHEMA_ADDITIONAL_PROPERTIES", None)
        if additional_props is not None:
            schema["additionalProperties"] = bool(additional_props)

        all_of = getattr(cls, "SCHEMA_ALL_OF", None)
        if all_of is not None:
            schema["allOf"] = copy.deepcopy(all_of)

        return schema

    def _build_demand_properties(self) -> Dict[str, Any]:
        types_mod = self._types
        base_properties = self._build_class_properties(types_mod.DemandConfig)
        base_properties.setdefault(_IMPORTS_KEY, self._imports_schema())
        base_properties.setdefault(_IMPORT_KEY, self._import_ref_schema())
        ordered: Dict[str, Any] = {}
        for name in types_mod.DEMAND_SCHEMA_PROPERTIES_ORDER:
            if name == "_templates":
                ordered[name] = {
                    "type": "object",
                    "description": "YAML anchor 模板集合(_templates), 供 fields/relations 复用",
                    "markdownDescription": (
                        "YAML anchor 模板集合.\n\n- 仅用于 YAML 复用(anchors)\n- 常用于 `fields` / `relations` / `retry`"
                    ),
                    "properties": {
                        "retry": {
                            "type": "object",
                            "description": "可复用的 retry 策略模板集合",
                            "markdownDescription": "可复用的 retry 策略模板集合.\n\n- key 为模板名\n- value 为 retry policy 对象",
                            "additionalProperties": {"$ref": "#/definitions/loader_retry"},
                        }
                    },
                    "additionalProperties": True,
                }
                continue
            if name == "fields":
                ordered[name] = {
                    "type": "object",
                    "description": "字段配置映射, key 为 field_id; 仅用于派生字段",
                    "markdownDescription": (
                        "字段配置映射(仅用于派生字段).\n\n"
                        "- 必须包含 `compute` 或 `call_by`\n"
                        "- 不能与源字段同名(避免 source/derived 重名)\n"
                        "- 支持 YAML anchor 复用"
                    ),
                    "additionalProperties": {"$ref": "#/definitions/field"},
                }
                continue
            ordered[name] = base_properties[name]
        return ordered

    def _build_field_definition(self) -> Dict[str, Any]:
        types_mod = self._types
        source_props = self._build_class_properties(types_mod.SourceFieldConfig)
        derived_props = self._build_class_properties(types_mod.DerivedFieldConfig)

        properties = dict(source_props)
        for name, schema in derived_props.items():
            if name in properties:
                if properties[name] != schema:
                    msg = "Field schema mismatch for '{}'".format(name)
                    raise ValueError(msg)
                continue
            properties[name] = schema

        properties.setdefault(_IMPORT_KEY, self._import_ref_schema())

        return {
            "type": "object",
            "additionalProperties": True,
            "properties": properties,
            "allOf": copy.deepcopy(types_mod.FIELD_DERIVED_CONDITIONS),
        }

    def _build_class_properties(self, cls: type) -> Dict[str, Any]:
        types_mod = self._types
        properties: Dict[str, Any] = {}
        for dc_field in dataclass_fields(cls):
            if dc_field.metadata.get(types_mod.SCHEMA_OMIT_KEY):
                continue
            meta = SchemaMeta.from_field(dc_field)
            prop_name = meta.schema_name or dc_field.name
            properties[prop_name] = self._build_field_schema(dc_field, meta)
        return properties

    def _build_field_schema(self, dc_field: "Field[Any]", meta: SchemaMeta) -> Dict[str, Any]:
        meta_payload = dict(meta.meta)
        if "ref" in meta_payload:
            ref_name = meta_payload.pop("ref")
            if not meta_payload:
                return {"$ref": "#/definitions/{}".format(ref_name)}
            expanded = self._expand_meta(meta_payload)
            expanded["allOf"] = [{"$ref": "#/definitions/{}".format(ref_name)}]
            return expanded
        if "schema" in meta_payload:
            schema = cast("Dict[str, Any]", copy.deepcopy(meta_payload.pop("schema")))
            schema.update(self._expand_meta(meta_payload))
            return schema

        schema = self._schema_for_type(dc_field.type)
        schema.update(self._expand_meta(meta_payload))
        return schema

    def schemas_equivalent(self, left: Any, right: Any) -> bool:
        return self.normalize_schema(left) == self.normalize_schema(right)

    def normalize_schema(self, value: Any, key: str = "") -> Any:
        if isinstance(value, dict):
            typed = cast("Dict[str, Any]", value)
            return {k: self.normalize_schema(v, k) for k, v in typed.items() if k not in self.IGNORED_KEYS}
        if isinstance(value, list):
            items = cast("List[Any]", value)
            normalized = [self.normalize_schema(item) for item in items]
            if key in self.ORDER_INSENSITIVE_KEYS:
                return sorted(normalized, key=self._sort_key)
            return normalized
        if isinstance(value, tuple):
            items = cast("Tuple[Any, ...]", value)
            return self.normalize_schema(list(items), key)
        return value

    def _sort_key(self, value: Any) -> str:
        return json.dumps(value, sort_keys=True, ensure_ascii=False, separators=(",", ":"))

    def _expand_meta(self, meta: Dict[str, Any]) -> Dict[str, Any]:
        expanded: Dict[str, Any] = {}
        items_choices = None
        for key, value in meta.items():
            mapped_key = self.META_KEY_MAP.get(key, key)
            if mapped_key == "additionalProperties":
                expanded[mapped_key] = self._expand_additional_props(value)
            elif mapped_key == "items_choices":
                items_choices = value
            else:
                expanded[mapped_key] = copy.deepcopy(value)

        if items_choices is not None:
            if "items" not in expanded:
                expanded["items"] = {}
            if isinstance(expanded["items"], dict):
                items_schema = cast("Dict[str, Any]", expanded["items"])
                items_schema["enum"] = copy.deepcopy(items_choices)

        if "examples" in expanded and not isinstance(expanded["examples"], list):
            expanded["examples"] = [expanded["examples"]]
        if "markdownDescription" not in expanded and "description" in expanded:
            expanded["markdownDescription"] = expanded["description"]

        return expanded

    def _expand_additional_props(self, value: Any) -> Any:
        if isinstance(value, str):
            return {"$ref": "#/definitions/{}".format(value)}
        return copy.deepcopy(value)

    def _schema_for_type(self, tp: Any) -> Dict[str, Any]:
        tp = self._strip_optional(tp)
        origin = getattr(tp, "__origin__", None)
        primitive = self._primitive_schema(tp)
        if primitive:
            return primitive

        container = self._container_schema(tp, origin)
        if container:
            return container

        ref_schema = self._ref_schema(tp)
        if ref_schema:
            return ref_schema

        return {}

    def _strip_optional(self, tp: Any) -> Any:
        origin = getattr(tp, "__origin__", None)
        if origin is Union:
            args = [arg for arg in getattr(tp, "__args__", ()) if arg is not type(None)]
            if len(args) == 1:
                return args[0]
        return tp

    def _primitive_schema(self, tp: Any) -> Dict[str, Any]:
        if isinstance(tp, type) and tp in self.PRIMITIVE_TYPE_MAP:
            return {"type": self.PRIMITIVE_TYPE_MAP[tp]}
        return {}

    def _container_schema(self, tp: Any, origin: Any) -> Dict[str, Any]:
        if origin is list or tp is list:
            args = getattr(tp, "__args__", ())
            item_type = args[0] if args else object
            return {"type": "array", "items": self._schema_for_type(item_type)}

        if origin is dict or tp is dict:
            return {"type": "object", "properties": {_IMPORT_KEY: self._import_ref_schema()}}

        if origin is tuple or tp is tuple:
            return self._tuple_schema(tp)

        return {}

    def _tuple_schema(self, tp: Any) -> Dict[str, Any]:
        args = cast("Tuple[Any, ...]", getattr(tp, "__args__", ()))
        if len(args) == self.ELLIPSIS_TUPLE_LEN and args[1] is Ellipsis:
            return {"type": "array", "items": self._schema_for_type(args[0])}

        if args:
            items = [self._schema_for_type(arg) for arg in args]
            return {
                "type": "array",
                "items": items,
                "minItems": len(items),
                "maxItems": len(items),
                "additionalItems": False,
            }

        return {"type": "array"}

    def _ref_schema(self, tp: Any) -> Dict[str, Any]:
        if isinstance(tp, type):
            schema_name = getattr(tp, "SCHEMA_NAME", None)
            if isinstance(schema_name, str):
                return {"$ref": "#/definitions/{}".format(schema_name)}
        return {}


_DEFAULT_BUILDER = SchemaBuilder()


def build_demand_schema() -> Dict[str, Any]:
    return _DEFAULT_BUILDER.build_demand_schema()


def build_workflow_schema() -> Dict[str, Any]:
    return _DEFAULT_BUILDER.build_workflow_schema()


def load_schema(path: Path) -> Dict[str, Any]:
    with path.open("r", encoding="utf-8") as handle:
        return json.load(handle)


def write_demand_schema(output_path: Path) -> None:
    schema = build_demand_schema()
    output_path.parent.mkdir(parents=True, exist_ok=True)
    with output_path.open("w", encoding="utf-8") as handle:
        json.dump(schema, handle, ensure_ascii=False, indent=2, sort_keys=False)
        _ = handle.write("\n")


def write_workflow_schema(output_path: Path) -> None:
    schema = build_workflow_schema()
    output_path.parent.mkdir(parents=True, exist_ok=True)
    with output_path.open("w", encoding="utf-8") as handle:
        json.dump(schema, handle, ensure_ascii=False, indent=2, sort_keys=False)
        _ = handle.write("\n")


def schemas_equivalent(left: Any, right: Any) -> bool:
    return _DEFAULT_BUILDER.schemas_equivalent(left, right)


def normalize_schema(value: Any, key: str = "") -> Any:
    return _DEFAULT_BUILDER.normalize_schema(value, key)



from dataclasses import dataclass
from dataclasses import field as dataclass_field
from typing import ClassVar, Optional, Tuple

from ..constants import (
    DEFAULT_PERF_REPORT_FORMAT,
    DEFAULT_PERF_SAMPLING_INTERVAL,
    DEFAULT_REL_LOG_TYPE_MISMATCH,
    DEFAULT_REL_REPORT_FORMAT,
    DEFAULT_RELATION_MAX_SAMPLES,
    DEFAULT_RELATION_SAMPLING_RATE,
    schema_meta,
)


@dataclass(frozen=True)
class PerformanceThresholdsConfig:
    SCHEMA_NAME: ClassVar[str] = "performance_thresholds"
    """性能阈值配置对象在 `YAML` 中的节点名称."""

    SCHEMA_ADDITIONAL_PROPERTIES: ClassVar[bool] = False
    """是否允许出现未声明的额外键."""

    batch_duration_warn: Optional[float] = dataclass_field(
        default=None,
        metadata=schema_meta(desc="批次耗时告警阈值(秒)"),
    )
    """批次耗时告警阈值(秒,可选)."""

    memory_increase_warn: Optional[float] = dataclass_field(
        default=None,
        metadata=schema_meta(desc="内存增长告警阈值(MB)"),
    )
    """内存增长告警阈值(兆字节,可选)."""


@dataclass(frozen=True)
class PerformanceReportConfig:
    SCHEMA_NAME: ClassVar[str] = "performance_report"
    """性能报告输出配置对象在 `YAML` 中的节点名称."""

    SCHEMA_ADDITIONAL_PROPERTIES: ClassVar[bool] = False
    """是否允许出现未声明的额外键."""

    format: str = dataclass_field(
        default=DEFAULT_PERF_REPORT_FORMAT,
        metadata=schema_meta(
            desc="报告输出格式",
            md=("报告输出格式.\n\n- `console`: 控制台输出\n- `json`: JSON 文件\n- `csv`: CSV 文件\n- `none`: 不输出报告"),
            choices=["console", "json", "csv", "none"],
            default=DEFAULT_PERF_REPORT_FORMAT,
            examples=["console"],
        ),
    )
    """报告输出格式:`console`/`json`/`csv`/`none`."""

    output: Optional[str] = dataclass_field(default=None, metadata=schema_meta(desc="报告输出路径"))
    """报告输出路径(可选)."""

    include_details: bool = dataclass_field(
        default=False,
        metadata=schema_meta(desc="包含详细统计", md="包含详细统计.", default=False),
    )
    """是否包含详细统计."""


@dataclass(frozen=True)
class PerformanceConfig:
    SCHEMA_NAME: ClassVar[str] = "performance"
    """性能可观测性配置对象在 `YAML` 中的节点名称."""

    SCHEMA_ADDITIONAL_PROPERTIES: ClassVar[bool] = False
    """是否允许出现未声明的额外键."""

    enabled: bool = dataclass_field(
        default=False,
        metadata=schema_meta(desc="启用性能监控", md="启用性能监控.", default=False),
    )
    """是否启用性能监控."""

    metrics: Tuple[str, ...] = dataclass_field(
        default=("duration",),
        metadata=schema_meta(
            desc="要收集的指标类型 (duration/memory/cpu)",
            md="要收集的指标类型.\n\n- 可选: duration / memory / cpu",
            items_choices=["duration", "memory", "cpu"],
        ),
    )
    """要收集的指标类型列表(例如 `duration`/`memory`/`cpu`)."""

    sampling_interval: int = dataclass_field(
        default=DEFAULT_PERF_SAMPLING_INTERVAL,
        metadata=schema_meta(
            desc="资源采样间隔(批次数)",
            md="资源采样间隔(按批次计).",
            min=1,
            default=DEFAULT_PERF_SAMPLING_INTERVAL,
        ),
    )
    """资源采样间隔(按批次数)."""

    report: Optional[PerformanceReportConfig] = dataclass_field(
        default=None,
        metadata=schema_meta(ref="performance_report"),
    )
    """性能报告输出配置(可选)."""

    thresholds: Optional[PerformanceThresholdsConfig] = dataclass_field(
        default=None,
        metadata=schema_meta(ref="performance_thresholds"),
    )
    """性能告警阈值配置(可选)."""


@dataclass(frozen=True)
class RelationReportConfig:
    SCHEMA_NAME: ClassVar[str] = "relation_report"
    """关联报告输出配置对象在 `YAML` 中的节点名称."""

    SCHEMA_ADDITIONAL_PROPERTIES: ClassVar[bool] = False
    """是否允许出现未声明的额外键."""

    format: str = dataclass_field(
        default=DEFAULT_REL_REPORT_FORMAT,
        metadata=schema_meta(
            desc="报告输出格式",
            md=("报告输出格式.\n\n- `console`: 控制台输出\n- `json`: JSON 文件\n- `none`: 不输出报告"),
            choices=["console", "json", "none"],
            default=DEFAULT_REL_REPORT_FORMAT,
            examples=["console"],
        ),
    )
    """报告输出格式:`console`/`json`/`none`."""

    output: Optional[str] = dataclass_field(default=None, metadata=schema_meta(desc="报告输出路径"))
    """报告输出路径(可选)."""


@dataclass(frozen=True)
class RelationsConfig:
    SCHEMA_NAME: ClassVar[str] = "relations"
    """关联可观测性配置对象在 `YAML` 中的节点名称."""

    SCHEMA_ADDITIONAL_PROPERTIES: ClassVar[bool] = False
    """是否允许出现未声明的额外键."""

    enabled: bool = dataclass_field(
        default=False,
        metadata=schema_meta(desc="启用关联可观测性", md="启用关联可观测性.", default=False),
    )
    """是否启用关联可观测性."""

    sampling_rate: float = dataclass_field(
        default=DEFAULT_RELATION_SAMPLING_RATE,
        metadata=schema_meta(
            desc="采样率(0.0-1.0)",
            md="采样率(0.0-1.0).",
            min=0.0,
            max=1.0,
            default=DEFAULT_RELATION_SAMPLING_RATE,
        ),
    )
    """采样率(0.0-1.0)."""

    log_type_mismatch: bool = dataclass_field(
        default=DEFAULT_REL_LOG_TYPE_MISMATCH,
        metadata=schema_meta(desc="记录类型不匹配日志", md="记录类型不匹配日志.", default=DEFAULT_REL_LOG_TYPE_MISMATCH),
    )
    """是否记录类型不匹配日志."""

    max_samples: int = dataclass_field(
        default=DEFAULT_RELATION_MAX_SAMPLES,
        metadata=schema_meta(
            desc="最大采样数量",
            md="最大采样数量.",
            min=0,
            default=DEFAULT_RELATION_MAX_SAMPLES,
        ),
    )
    """最大采样数量."""

    report: Optional[RelationReportConfig] = dataclass_field(
        default=None,
        metadata=schema_meta(ref="relation_report"),
    )
    """关联报告输出配置(可选)."""


@dataclass(frozen=True)
class VizConfig:
    SCHEMA_NAME: ClassVar[str] = "viz"
    """可视化输出配置对象在 `YAML` 中的节点名称."""

    SCHEMA_ADDITIONAL_PROPERTIES: ClassVar[bool] = False
    """是否允许出现未声明的额外键."""

    enabled: bool = dataclass_field(
        default=False,
        metadata=schema_meta(desc="启用 Scalim Viz 输出", md="启用 Scalim Viz 输出.", default=False),
    )
    """是否启用 `Scalim Viz` 输出."""

    output_dir: Optional[str] = dataclass_field(
        default=None,
        metadata=schema_meta(
            desc="输出目录(自动追加 scalim-viz)",
            md=(
                "输出目录(自动追加 scalim-viz).\n\n"
                "- 写入路径: output_dir/scalim-viz/<run_id>/viz_*.json\n"
                "- 若 output_dir 已包含 scalim-viz,则直接使用\n"
                "- 若显式配置 output_path/snapshot_path,则忽略该规则"
            ),
        ),
    )
    """输出目录(会自动追加 `scalim-viz`,可选)."""

    output_path: Optional[str] = dataclass_field(
        default=None,
        metadata=schema_meta(desc="事件输出文件路径(可选)", md="事件输出文件路径(可选)."),
    )
    """事件输出文件路径(可选)."""

    snapshot_path: Optional[str] = dataclass_field(
        default=None,
        metadata=schema_meta(desc="快照输出文件路径(可选)", md="快照输出文件路径(可选)."),
    )
    """快照输出文件路径(可选)."""

    append: bool = dataclass_field(
        default=False,
        metadata=schema_meta(
            desc="事件文件追加写入",
            md=(
                "事件文件追加写入(可选).\n\n"
                "- 默认 `false`: 每次运行会覆盖 `output_path` 对应文件,避免跨 run 混写\n"
                "- 设为 `true` 时,将以 JSONL 追加写入(旧行为)\n"
                "- `output_dir` 写入到按 run 隔离目录时,通常不需要开启"
            ),
            default=False,
            examples=[False],
        ),
    )
    """是否以 `JSONL` 追加写入事件文件."""

    trace_enabled: bool = dataclass_field(
        default=False,
        metadata=schema_meta(
            desc="启用高频 trace 输出",
            md=(
                "启用高频 trace 输出(可选).\n\n"
                "- `false`(默认): 仅输出编排级事件到 `viz_events.jsonl`\n"
                "- `true`: 额外输出高频 trace 事件到 `viz_trace.jsonl`(例如 field/row/relation lookup)\n\n"
                "建议: 默认关闭;需要深挖时在 UI 勾选加载 trace 并配合过滤/步进 lens 使用.\n\n"
                "旧字段 `observability.viz.event_mode` 已移除,请使用 `trace_enabled`."
            ),
            default=False,
            examples=[False],
        ),
    )
    """是否启用高频 `trace` 事件输出."""

    payload_policy: str = dataclass_field(
        default="summary",
        metadata=schema_meta(
            desc="事件 payload 策略",
            md=("事件 payload 策略.\n\n- `none`: 不输出 payload\n- `summary`: 仅摘要\n- `sample`: 抽样\n- `full`: 全量"),
            choices=["none", "summary", "sample", "full"],
            default="summary",
            examples=["summary"],
        ),
    )
    """事件负载策略:`none`/`summary`/`sample`/`full`."""

    sample_size: int = dataclass_field(
        default=5,
        metadata=schema_meta(
            desc="sample 策略下的样本数量",
            md="sample 策略下的样本数量.",
            min=0,
            default=5,
        ),
    )
    """在 `sample` 策略下的样本数量."""

    run_name: Optional[str] = dataclass_field(
        default=None,
        metadata=schema_meta(desc="运行名称", md="运行名称(可选)."),
    )
    """运行名称(可选)."""

    env: Optional[str] = dataclass_field(
        default=None,
        metadata=schema_meta(desc="运行环境标签", md="运行环境标签(可选)."),
    )
    """运行环境标签(可选)."""

    use_default_output_dir: bool = dataclass_field(
        default=False,
        metadata=schema_meta(desc="使用默认输出目录", md="使用默认输出目录(~/.config/scalim-viz 等).", default=False),
    )
    """是否使用默认输出目录(例如 `~/.config/scalim-viz`)."""


@dataclass(frozen=True)
class LoggingConfig:
    SCHEMA_NAME: ClassVar[str] = "logging"
    """日志观测配置对象在 `YAML` 中的节点名称."""

    SCHEMA_ADDITIONAL_PROPERTIES: ClassVar[bool] = False
    """是否允许出现未声明的额外键."""

    enabled: bool = dataclass_field(
        default=True,
        metadata=schema_meta(desc="启用日志观测", md="启用日志观测.", default=True),
    )
    """是否启用日志观测."""

    renderer: str = dataclass_field(
        default="pretty",
        metadata=schema_meta(
            desc="日志渲染器(pretty/logger)",
            md="日志渲染器.\n\n- `pretty`: 输出到 pretty console(如 panel/table)\n- `logger`: 输出到标准 logger",
            choices=["pretty", "logger"],
            default="pretty",
            examples=["pretty"],
        ),
    )
    """日志渲染器:`pretty` 或 `logger`."""


@dataclass(frozen=True)
class TraceConfig:
    SCHEMA_NAME: ClassVar[str] = "trace"
    """执行追踪配置对象在 `YAML` 中的节点名称."""

    SCHEMA_ADDITIONAL_PROPERTIES: ClassVar[bool] = False
    """是否允许出现未声明的额外键."""

    enabled: bool = dataclass_field(
        default=False,
        metadata=schema_meta(desc="启用执行追踪", md="启用执行追踪.", default=False),
    )
    """是否启用执行追踪."""


@dataclass(frozen=True)
class RowGapConfig:
    SCHEMA_NAME: ClassVar[str] = "row_gap"
    """行缺口统计配置对象在 `YAML` 中的节点名称."""

    SCHEMA_ADDITIONAL_PROPERTIES: ClassVar[bool] = False
    """是否允许出现未声明的额外键."""

    enabled: bool = dataclass_field(
        default=False,
        metadata=schema_meta(desc="启用行缺口统计", md="启用行缺口统计.", default=False),
    )
    """是否启用行缺口统计."""

    primary_loader_name: str = dataclass_field(
        default="primary_keys",
        metadata=schema_meta(desc="主数据 loader 名称", md="主数据 loader 名称.", default="primary_keys"),
    )
    """主数据加载器名称."""

    data_loader_names: Tuple[str, ...] = dataclass_field(
        default=("base_info",),
        metadata=schema_meta(
            desc="参与统计的 loader 列表",
            md="参与统计的 loader 列表.",
            items={"type": "string"},
        ),
    )
    """参与统计的加载器名称列表."""

    sample_limit: int = dataclass_field(
        default=5,
        metadata=schema_meta(desc="缺口采样数量", md="缺口采样数量.", min=0, default=5),
    )
    """缺口采样数量上限."""


@dataclass(frozen=True)
class MemoryOptimizationConfig:
    SCHEMA_NAME: ClassVar[str] = "memory_opt"
    """内存优化统计配置对象在 `YAML` 中的节点名称."""

    SCHEMA_ADDITIONAL_PROPERTIES: ClassVar[bool] = False
    """是否允许出现未声明的额外键."""

    enabled: bool = dataclass_field(
        default=False,
        metadata=schema_meta(desc="启用内存优化统计", md="启用内存优化统计.", default=False),
    )
    """是否启用内存优化统计."""

    auto_report: bool = dataclass_field(
        default=False,
        metadata=schema_meta(desc="自动输出摘要", md="自动输出摘要.", default=False),
    )
    """是否自动输出摘要."""

    max_fields: int = dataclass_field(
        default=0,
        metadata=schema_meta(desc="摘要字段上限", md="摘要字段上限(0 表示不限制).", min=0, default=0),
    )
    """摘要字段上限(`0` 表示不限制)."""


@dataclass(frozen=True)
class ObservabilityConfig:
    SCHEMA_NAME: ClassVar[str] = "observability"
    """可观测性配置对象在 `YAML` 中的节点名称."""

    SCHEMA_ADDITIONAL_PROPERTIES: ClassVar[bool] = False
    """是否允许出现未声明的额外键."""

    logging: Optional[LoggingConfig] = dataclass_field(
        default=None,
        metadata=schema_meta(
            desc="日志观测配置",
            md="日志观测配置.\n\n- 控制日志输出开启/关闭",
            ref="logging",
        ),
    )
    """日志观测配置(可选)."""

    performance: Optional[PerformanceConfig] = dataclass_field(
        default=None,
        metadata=schema_meta(
            desc="性能可观测性配置",
            md="性能可观测性配置.\n\n- 关注耗时/内存/CPU 等指标",
            ref="performance",
        ),
    )
    """性能可观测性配置(可选)."""

    relations: Optional[RelationsConfig] = dataclass_field(
        default=None,
        metadata=schema_meta(
            desc="关联可观测性配置",
            md="关联可观测性配置.\n\n- 关注关联步骤的采样与类型校验",
            ref="relations",
        ),
    )
    """关联可观测性配置(可选)."""

    viz: Optional[VizConfig] = dataclass_field(
        default=None,
        metadata=schema_meta(
            desc="Scalim Viz 可视化输出配置",
            md="Scalim Viz 可视化输出配置.\n\n- 输出 viz_snapshot.json + viz_events.jsonl",
            ref="viz",
        ),
    )
    """可视化输出配置(可选)."""

    trace: Optional[TraceConfig] = dataclass_field(
        default=None,
        metadata=schema_meta(
            desc="执行追踪配置",
            md="执行追踪配置.\n\n- 记录批次级执行步骤",
            ref="trace",
        ),
    )
    """执行追踪配置(可选)."""

    row_gap: Optional[RowGapConfig] = dataclass_field(
        default=None,
        metadata=schema_meta(
            desc="行缺口统计配置",
            md="行缺口统计配置.\n\n- 统计 loader 期望/实际行数差异",
            ref="row_gap",
        ),
    )
    """行缺口统计配置(可选)."""

    memory_opt: Optional[MemoryOptimizationConfig] = dataclass_field(
        default=None,
        metadata=schema_meta(
            desc="内存优化统计配置",
            md="内存优化统计配置.\n\n- 汇总字段瘦身/行释放等事件",
            ref="memory_opt",
        ),
    )
    """内存优化统计配置(可选)."""



from typing import Callable, Tuple, TypeVar

HookT = TypeVar("HookT")
EventT = TypeVar("EventT")
HandlerResultT = TypeVar("HandlerResultT")


class HookDispatchStrategy:
    """分发 `hook` 处理器;可插拔以定制事件扇出策略."""

    def dispatch(
        self,
        handler_pairs: Tuple[Tuple[HookT, Callable[[EventT], HandlerResultT]], ...],
        event: EventT,
        safe_call: Callable[[HookT, Callable[[EventT], HandlerResultT], EventT], None],
    ) -> None:
        for hook, handler in handler_pairs:
            safe_call(hook, handler, event)


__all__ = ["HookDispatchStrategy"]



import csv
import logging
from pathlib import Path
from typing import Optional

from ...vendor.literich import Table
from ..perf_metrics import PerformanceMetrics


class PerformancePresentationLayer:
    """性能指标展示/导出层."""

    def output_report(
        self,
        *,
        metrics: PerformanceMetrics,
        report_format: str,
        output_path: Optional[str],
        include_details: bool,
        logger: logging.Logger,
    ) -> None:
        if report_format == "none":
            return
        if report_format == "console":
            logger.info(self.render_summary(metrics, include_details=include_details))
            return
        if report_format == "json":
            self.write_json_report(metrics=metrics, output_path=output_path, logger=logger)
            return
        if report_format == "csv":
            self.write_csv_report(metrics=metrics, output_path=output_path, logger=logger)

    def write_json_report(
        self,
        *,
        metrics: PerformanceMetrics,
        output_path: Optional[str],
        logger: logging.Logger,
    ) -> None:
        if not output_path:
            logger.info("\n%s", metrics.to_json())
            return

        try:
            path = Path(output_path)
            if path.parent and not path.parent.exists():
                path.parent.mkdir(parents=True, exist_ok=True)
            _ = path.write_text(metrics.to_json(), encoding="utf-8")
            logger.info("[PerformanceObserver] 报告已写入: %s", output_path)
        except OSError as e:
            logger.warning("[PerformanceObserver] 写入报告失败: %s", e)

    def write_csv_report(
        self,
        *,
        metrics: PerformanceMetrics,
        output_path: Optional[str],
        logger: logging.Logger,
    ) -> None:
        if not output_path:
            logger.warning("[PerformanceObserver] `CSV` 输出需要提供 output_path")
            return

        try:
            rows = metrics.to_csv_rows()
            if not rows:
                return

            path = Path(output_path)
            if path.parent and not path.parent.exists():
                path.parent.mkdir(parents=True, exist_ok=True)

            with path.open("w", newline="", encoding="utf-8") as f:
                writer = csv.DictWriter(f, fieldnames=list(rows[0].keys()))
                writer.writeheader()
                writer.writerows(rows)

            logger.info("[PerformanceObserver] 报告已写入: %s", output_path)
        except OSError as e:
            logger.warning("[PerformanceObserver] 写入报告失败: %s", e)

    def render_summary(self, metrics: PerformanceMetrics, *, include_details: bool) -> str:
        summary_table = Table(title="Performance Summary", border_style="box")
        _ = summary_table.add_column("Metric", min_width=20)
        _ = summary_table.add_column("Value", min_width=15, align="right")

        _ = summary_table.add_row("Total Duration", "{:.3f}s".format(metrics.total_duration))
        _ = summary_table.add_row("Input Rows (row_ids)", str(metrics.total_rows))
        _ = summary_table.add_row("Throughput (row_ids/s)", "{:.1f} rows/s".format(metrics.throughput))
        _ = summary_table.add_row("Batch Count", str(metrics.batch_count))
        _ = summary_table.add_row("Avg Batch Duration", "{:.4f}s".format(metrics.avg_batch_duration))

        if metrics.peak_memory_mb is not None:
            _ = summary_table.add_row("Peak Memory", "{:.1f} MB".format(metrics.peak_memory_mb))
        if metrics.memory_increase_mb is not None:
            _ = summary_table.add_row("Memory Increase", "{:.1f} MB".format(metrics.memory_increase_mb))

        output_lines = ["\n" + summary_table.render()]

        stages = metrics.stage_metrics
        if stages.loader_duration > 0 or stages.compute_duration > 0 or stages.write_duration > 0:
            stage_table = Table(title="Stage Breakdown", border_style="box")
            _ = stage_table.add_column("Stage", min_width=12)
            _ = stage_table.add_column("Duration", min_width=12, align="right")
            _ = stage_table.add_column("Percent", min_width=10, align="right")

            total = stages.loader_duration + stages.compute_duration + stages.write_duration
            if total > 0:
                _ = stage_table.add_row(
                    "Loader",
                    "{:.3f}s".format(stages.loader_duration),
                    "{:.1f}%".format(100 * stages.loader_duration / total),
                )
                _ = stage_table.add_row(
                    "Compute",
                    "{:.3f}s".format(stages.compute_duration),
                    "{:.1f}%".format(100 * stages.compute_duration / total),
                )
                _ = stage_table.add_row(
                    "Write",
                    "{:.3f}s".format(stages.write_duration),
                    "{:.1f}%".format(100 * stages.write_duration / total),
                )
                output_lines.append(stage_table.render())

        if metrics.loader_stats and include_details:
            loader_table = Table(title="Loader Statistics", border_style="box")
            _ = loader_table.add_column("Loader", min_width=20)
            _ = loader_table.add_column("Calls", min_width=6, align="right")
            _ = loader_table.add_column("Records", min_width=8, align="right")
            _ = loader_table.add_column("Avg Time", min_width=10, align="right")

            for name, stats in metrics.loader_stats.items():
                _ = loader_table.add_row(
                    name,
                    str(stats.call_count),
                    str(stats.total_records),
                    "{:.4f}s".format(stats.avg_duration),
                )
            output_lines.append(loader_table.render())

        return "\n".join(output_lines)


__all__ = ["PerformancePresentationLayer"]



# region imports

import unicodedata
from typing import Any, Callable, List, Optional

from ..compact.typing_extensionsx import override

# endregion

TextAlign = str  # 对齐方式:`"left"` | `"center"` | `"right"`


def _display_width(text: str) -> int:
    width = 0
    for char in text:
        if unicodedata.east_asian_width(char) in ("F", "W"):
            width += 2
        else:
            width += 1
    return width


def _pad_cell(text: str, width: int, align: TextAlign = "left") -> str:
    text_width = _display_width(text)
    padding = width - text_width
    if padding <= 0:
        return text
    if align == "center":
        left_pad = padding // 2
        right_pad = padding - left_pad
        return " " * left_pad + text + " " * right_pad
    if align == "right":
        return " " * padding + text
    return text + " " * padding


class Column:
    header: str
    min_width: int
    max_width: int
    align: TextAlign
    formatter: Optional[Callable[[Any], str]]

    def __init__(
        self,
        header: str,
        *,
        min_width: int = 0,
        max_width: int = 50,
        align: TextAlign = "left",
        formatter: Optional[Callable[[Any], str]] = None,
    ) -> None:
        self.header = header
        self.min_width = max(min_width, _display_width(header))
        self.max_width = max_width
        self.align = align
        self.formatter = formatter

    def format_value(self, value: Any) -> str:
        if self.formatter:
            return self.formatter(value)
        if value is None:
            return ""
        return str(value)


class Table:
    title: Optional[str]
    columns: List[Column]
    rows: List[List[Any]]
    show_header: bool
    border_style: str  # 边框样式:`"simple"` | `"box"` | `"none"`
    _col_widths: List[int]

    def __init__(
        self,
        title: Optional[str] = None,
        *,
        show_header: bool = True,
        border_style: str = "box",
    ) -> None:
        self.title = title
        self.columns = []
        self.rows = []
        self.show_header = show_header
        self.border_style = border_style
        self._col_widths = []

    def add_column(
        self,
        header: str,
        *,
        min_width: int = 0,
        max_width: int = 50,
        align: TextAlign = "left",
        formatter: Optional[Callable[[Any], str]] = None,
    ) -> "Table":
        col = Column(
            header,
            min_width=min_width,
            max_width=max_width,
            align=align,
            formatter=formatter,
        )
        self.columns.append(col)
        return self

    def add_row(self, *values: Any) -> "Table":
        self.rows.append(list(values))
        return self

    def _compute_widths(self) -> None:
        self._col_widths = [col.min_width for col in self.columns]

        for col_idx, col in enumerate(self.columns):
            header_width = _display_width(col.header)
            if header_width > self._col_widths[col_idx]:
                self._col_widths[col_idx] = min(header_width, col.max_width)

        for row in self.rows:
            for col_idx, value in enumerate(row):
                if col_idx >= len(self.columns):
                    break
                col = self.columns[col_idx]
                cell_text = col.format_value(value)
                cell_width = _display_width(cell_text)
                if cell_width > self._col_widths[col_idx]:
                    self._col_widths[col_idx] = min(cell_width, col.max_width)

    def _truncate(self, text: str, width: int) -> str:
        if _display_width(text) <= width:
            return text
        if width <= 3:  # noqa: PLR2004
            return text[:width]
        result = ""
        current_width = 0
        for char in text:
            char_width = 2 if unicodedata.east_asian_width(char) in ("F", "W") else 1
            if current_width + char_width > width - 2:
                break
            result += char
            current_width += char_width
        return result + ".."

    def _render_box(self) -> str:  # noqa: C901
        self._compute_widths()
        lines: List[str] = []

        total_width = sum(self._col_widths) + len(self._col_widths) * 3 + 1

        if self.title:
            lines.append("┌" + "─" * (total_width - 2) + "┐")
            title_padded = _pad_cell(self.title, total_width - 4, "center")
            lines.append("│ " + title_padded + " │")
            lines.append("├" + "─" * (total_width - 2) + "┤")
        else:
            top_border = "┌"
            for i, w in enumerate(self._col_widths):
                top_border += "─" * (w + 2)
                if i < len(self._col_widths) - 1:
                    top_border += "┬"
            top_border += "┐"
            lines.append(top_border)

        if self.show_header:
            header_cells: List[str] = []
            for col_idx, col in enumerate(self.columns):
                w = self._col_widths[col_idx]
                cell = _pad_cell(self._truncate(col.header, w), w, col.align)
                header_cells.append(cell)
            lines.append("│ " + " │ ".join(header_cells) + " │")

            sep = "├"
            for i, w in enumerate(self._col_widths):
                sep += "─" * (w + 2)
                if i < len(self._col_widths) - 1:
                    sep += "┼"
            sep += "┤"
            lines.append(sep)

        for row in self.rows:
            row_cells: List[str] = []
            for col_idx, col in enumerate(self.columns):
                value = row[col_idx] if col_idx < len(row) else None
                w = self._col_widths[col_idx]
                cell_text = col.format_value(value)
                cell = _pad_cell(self._truncate(cell_text, w), w, col.align)
                row_cells.append(cell)
            lines.append("│ " + " │ ".join(row_cells) + " │")

        bottom_border = "└"
        for i, w in enumerate(self._col_widths):
            bottom_border += "─" * (w + 2)
            if i < len(self._col_widths) - 1:
                bottom_border += "┴"
        bottom_border += "┘"
        lines.append(bottom_border)

        return "\n".join(lines)

    def _render_simple(self) -> str:
        self._compute_widths()
        lines: List[str] = []

        if self.title:
            lines.append(self.title)
            lines.append("-" * _display_width(self.title))

        if self.show_header:
            header_cells: List[str] = []
            for col_idx, col in enumerate(self.columns):
                w = self._col_widths[col_idx]
                cell = _pad_cell(self._truncate(col.header, w), w, col.align)
                header_cells.append(cell)
            lines.append("  ".join(header_cells))
            lines.append("  ".join("-" * w for w in self._col_widths))

        for row in self.rows:
            row_cells: List[str] = []
            for col_idx, col in enumerate(self.columns):
                value = row[col_idx] if col_idx < len(row) else None
                w = self._col_widths[col_idx]
                cell_text = col.format_value(value)
                cell = _pad_cell(self._truncate(cell_text, w), w, col.align)
                row_cells.append(cell)
            lines.append("  ".join(row_cells))

        return "\n".join(lines)

    def _render_none(self) -> str:
        self._compute_widths()
        lines: List[str] = []

        if self.title:
            lines.append(self.title)

        if self.show_header:
            header_cells: List[str] = []
            for col_idx, col in enumerate(self.columns):
                w = self._col_widths[col_idx]
                cell = _pad_cell(self._truncate(col.header, w), w, col.align)
                header_cells.append(cell)
            lines.append(" ".join(header_cells))

        for row in self.rows:
            row_cells: List[str] = []
            for col_idx, col in enumerate(self.columns):
                value = row[col_idx] if col_idx < len(row) else None
                w = self._col_widths[col_idx]
                cell_text = col.format_value(value)
                cell = _pad_cell(self._truncate(cell_text, w), w, col.align)
                row_cells.append(cell)
            lines.append(" ".join(row_cells))

        return "\n".join(lines)

    def render(self) -> str:
        if self.border_style == "simple":
            return self._render_simple()
        if self.border_style == "none":
            return self._render_none()
        return self._render_box()

    @override
    def __str__(self) -> str:
        return self.render()


class Panel:
    title: str
    content: str
    width: int
    padding: int

    def __init__(
        self,
        content: str,
        *,
        title: str = "",
        width: int = 60,
        padding: int = 1,
    ) -> None:
        self.content = content
        self.title = title
        self.width = width
        self.padding = padding

    def render(self) -> str:
        lines: List[str] = []
        inner_width = self.width - 2

        lines.append("┌" + "─" * inner_width + "┐")

        if self.title:
            title_padded = _pad_cell(self.title, inner_width, "center")
            lines.append("│" + title_padded + "│")
            lines.append("├" + "─" * inner_width + "┤")

        pad = " " * self.padding
        for line in self.content.split("\n"):
            text_width = inner_width - self.padding * 2
            text_padded = _pad_cell(line, text_width, "left")
            lines.append("│" + pad + text_padded + pad + "│")

        lines.append("└" + "─" * inner_width + "┘")

        return "\n".join(lines)

    @override
    def __str__(self) -> str:
        return self.render()


__all__ = [
    "Column",
    "Panel",
    "Table",
]



from ....utils.relation_signature import (
    LoadRefCacheKey,
    LookupCastSignature,
    RelationSignature,
    StepSignature,
    build_binding_signature,
    build_relation_signature,
    build_step_signature,
    can_group_by_relation,
    has_rows_binding,
    is_auto_lookup_cast,
    lookup_cast_signature,
    normalize_key_field,
    resolve_step_binding,
)

__all__ = [
    "LoadRefCacheKey",
    "LookupCastSignature",
    "RelationSignature",
    "StepSignature",
    "build_binding_signature",
    "build_relation_signature",
    "build_step_signature",
    "can_group_by_relation",
    "has_rows_binding",
    "is_auto_lookup_cast",
    "lookup_cast_signature",
    "normalize_key_field",
    "resolve_step_binding",
]
